"""Defines several methods for analyzing, plotting, and exporting wearable data, including a Pandas accessor for wearable dataframes"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/05_readers.ipynb.

# %% auto 0
__all__ = ['VALID_WEARABLE_STREAMS', 'ACTIWATCH_COLUMN_RENAMING', 'WEARABLE_RESAMPLE_METHOD', 'WearableData', 'load_json',
           'load_csv', 'load_actiwatch', 'interval_fraction', 'resample_df', 'combine_wearable_dataframes']

# %% ../nbs/api/05_readers.ipynb 4
import json
import numpy as np
import pandas as pd
from typing import Dict

# %% ../nbs/api/05_readers.ipynb 6
VALID_WEARABLE_STREAMS = ['steps', 'heartrate', 'wake', 'light_estimate', 'activity']

# %% ../nbs/api/05_readers.ipynb 7
@pd.api.extensions.register_dataframe_accessor("wearable")
class WearableData:
    "pd.DataFrame accessor implementing wearable-specific methods"
    def __init__(self, pandas_obj):
        self._validate_columns(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate_columns(obj):
        if 'datetime' not in obj.columns:
            if 'start' not in obj.columns and 'end' not in obj.columns:
                raise AttributeError("DataFrame must have 'datetime' column or 'start' and 'end' columns")

        if not any([col in obj.columns for col in VALID_WEARABLE_STREAMS]):
            raise AttributeError(f"DataFrame must have at least one wearable data column from: {VALID_WEARABLE_STREAMS}.")
        
    @staticmethod
    def _validate_metadata(metadata):
        if metadata:
            if not isinstance(metadata, dict):
                raise AttributeError("Metadata must be a dictionary.")
            if not any([key in metadata.keys() for key in ['data_id', 'subject_id']]):
                raise AttributeError("Metadata must have at least one of the following keys: data_id, subject_id.")
            if not all([isinstance(value, str) for value in metadata.values()]):
                raise AttributeError("Metadata values must be strings.")
    
    @staticmethod
    def rename_columns(df, 
                       inplace: bool = False
                       ):
        "Standardize column names by making them lowercase and replacing spaces with underscores"
        columns = [col.lower().replace(' ', '_') for col in df.columns]
        if inplace:
            df.columns = columns
        else:
            new_df = df.copy()
            new_df.columns = columns
            return new_df

    def is_valid(self):
        self._validate_columns(self._obj)
        self._validate_metadata(self._obj.attrs)
        return True

    def add_metadata(self,
                     metadata: Dict[str, str], # metadata containing data_id, subject_id, or other_info
                     inplace: bool = False, # whether to return a new dataframe or modify the current one
                     ):
        self._validate_metadata(metadata)
        if inplace:
            for key, value in metadata.items():
                self._obj.attrs[key] = value
        else:
            obj = self._obj.copy()
            for key, value in metadata.items():
                obj.attrs[key] = value
            return obj

# %% ../nbs/api/05_readers.ipynb 9
def load_json(filepath: str, # path to file
              metadata: Dict[str, str] = None, # metadata containing data_id, subject_id, or other_info
              ) -> Dict[str, pd.DataFrame]: # dictionary of wearable dataframes, one key:value pair per wearable data stream
    "Create a dataframe from a json containing a single or multiple streams of wearable data"
    # validate inputs
    if not isinstance(filepath, str):
        raise AttributeError("Filepath must be a string.")
    if metadata is not None:
        WearableData._validate_metadata(metadata)
    # load json
    jdict = json.load(open(filepath, 'r'))
    # check that it contains valid keys
    if not np.all([key in VALID_WEARABLE_STREAMS for key in jdict.keys()]):
        raise AttributeError("Invalid keys in JSON file. At least one key must be steps, heartrate, wake, light_estimate, or activity.")
    # create a df for each wearable stream
    df_dict = {}
    for key in jdict.keys():
        if key in VALID_WEARABLE_STREAMS:
            df_dict[key] = pd.DataFrame.from_dict(jdict[key])
        else:
            print(f"Excluded key: {key} because it's not a valid wearable stream column name.")
    for key in df_dict.keys():
        df = df_dict[key]
        if 'timestamp' in df.columns:
            df['datetime'] = pd.to_datetime(df['timestamp'], unit='s')
        elif 'start' in df.columns and 'end' in df.columns:
            df['start'] = pd.to_datetime(df['start'], unit='s')
            df['end'] = pd.to_datetime(df['end'], unit='s')
        if metadata is not None:
            df.wearable.add_metadata(metadata, inplace=True)
        else:
            df.wearable.add_metadata({'data_id': 'unknown', 'subject_id': 'unknown'}, inplace=True)
        df_dict[key] = df
    return df_dict

# %% ../nbs/api/05_readers.ipynb 10
def load_csv(filepath: str, # full path to csv file to be loaded
             metadata: Dict[str, str] = None, # metadata containing data_id, subject_id, or other_info
             timestamp_col: str = None, # name of the column to be used as timestamp. If None, it is assumed that a `datetime` column exists
             *args, # arguments to pass to pd.read_csv
             **kwargs, # keyword arguments to pass to pd.read_csv
             ):
    "Create a dataframe from a csv containing wearable data"
    # validate inputs
    if not isinstance(filepath, str):
        raise AttributeError("Filepath must be a string.")
    if not isinstance(timestamp_col, str) and timestamp_col is not None:
        raise AttributeError("Timestamp column must be a string.")
    if metadata is not None:
        WearableData._validate_metadata(metadata)
    # load csv
    df = pd.read_csv(filepath, *args, **kwargs)
    # create datetime column
    if timestamp_col is not None:
        df['datetime'] = pd.to_datetime(df[timestamp_col], unit='s')
    if timestamp_col is None:
        if 'datetime' in df.columns:
            df['datetime'] = pd.to_datetime(df['datetime'])
        elif 'start' in df.columns and 'end' in df.columns:
            df['start'] = pd.to_datetime(df['start'])
            df['end'] = pd.to_datetime(df['end'])
        if 'datetime' not in df.columns and 'start' not in df.columns and 'end' not in df.columns:
            raise AttributeError("CSV file must have a column named 'datetime' or 'start' and 'end'")
    # add metadata
    if metadata is not None:
        df.wearable.add_metadata(metadata, inplace=True)
    else:
        df.wearable.add_metadata({'data_id': 'unknown', 'subject_id': 'unknown'}, inplace=True)
    return df

# %% ../nbs/api/05_readers.ipynb 11
ACTIWATCH_COLUMN_RENAMING = {
    'White Light': 'light_estimate',
    'Sleep/Wake': 'wake',
    'Activity': 'activity',
} 

# %% ../nbs/api/05_readers.ipynb 12
def load_actiwatch(filepath: str, # full path to csv file to be loaded
                   metadata: Dict[str, str] = None, # metadata containing data_id, subject_id, or other_info
                   *args, # arguments to pass to pd.read_csv
                   **kwargs, # keyword arguments to pass to pd.read_csv
                   ) -> pd.DataFrame: # dataframe with the wearable data
    "Create a dataframe from an actiwatch csv file"
    # validate inputs
    if not isinstance(filepath, str):
        raise AttributeError("Filepath must be a string.")
    if metadata is not None:
        WearableData._validate_metadata(metadata)
    # load csv
    df = pd.read_csv(filepath, *args, **kwargs)
    df['datetime'] = pd.to_datetime(df['Date']+" "+df['Time'])
    # drop unnecessary columns
    df.drop(columns=['Date', 'Time'], inplace=True)
    # rename columns
    df.rename(columns=ACTIWATCH_COLUMN_RENAMING, inplace=True)
    # add metadata
    if metadata is not None:
        df.wearable.add_metadata(metadata, inplace=True)
    else:
        df.wearable.add_metadata({'data_id': 'unknown', 'subject_id': 'unknown'}, inplace=True)
    return df

# %% ../nbs/api/05_readers.ipynb 14
def interval_fraction(
        starts: pd.Series, # start datetimes of intervals
        stops: pd.Series, # stop datetimes of intervals
        ref_start: pd.Timestamp, # start datetime of reference interval
        ref_stop: pd.Timestamp # stop datetime of reference interval
        ):
        "Calculate the fraction of each interval contained in the reference interval."
        max_starts = starts.apply(lambda x: max(x, ref_start))
        min_ends = stops.apply(lambda x: min(x, ref_stop))
        contained_intervals = (min_ends - max_starts).apply(lambda x: x.seconds)
        full_intervals = (stops - starts).apply(lambda x: x.seconds)
        return contained_intervals / full_intervals

# %% ../nbs/api/05_readers.ipynb 15
def resample_df(df: pd.DataFrame, # dataframe to be resampled
                name: str, # name of the wearable data to resample (one of steps, heartrate, wake, light_estimate, or activity)
                freq: str, # frequency to resample to. String must be a valid pandas frequency string (e.g. '1min', '5min', '1H', '1D'). See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
                agg_method: str, # aggregation method to use when resampling
                initial_datetime: pd.Timestamp = None, # initial datetime to use when resampling. If None, the minimum datetime in the dataframe is used
                final_datetime: pd.Timestamp = None, # final datetime to use when resampling. If None, the maximum datetime in the dataframe is used
                ) -> pd.DataFrame: # resampled dataframe
    "Resample a wearable dataframe. If data is specified in intervals, returns the density of the quantity per minute."
    # validate inputs
    if not df.wearable.is_valid():
        raise AttributeError("Dataframe must be a valid wearable dataframe.")
    if not isinstance(df, pd.DataFrame):
        raise AttributeError("Dataframe must be a pandas dataframe.")
    if not isinstance(freq, str):
        raise AttributeError("Frequency must be a string.")
    if name is not None and name not in VALID_WEARABLE_STREAMS:
        raise AttributeError(f"Name must be one of: {VALID_WEARABLE_STREAMS}.")
    if name not in df.columns:
        raise AttributeError(f"Name must be one of: {df.columns}.")
    if agg_method not in ['sum', 'mean', 'max', 'min']:
        raise AttributeError("Aggregation method must be one of: sum, mean, max, min.")
    if initial_datetime is not None and not isinstance(initial_datetime, pd.Timestamp):
        raise AttributeError("Initial datetime must be a pandas timestamp.")
    if final_datetime is not None and not isinstance(final_datetime, pd.Timestamp):
        raise AttributeError("Final datetime must be a pandas timestamp.")
    # resample
    values = df[name]
    if 'start' in df.columns and 'end' in df.columns:
        # data is specified in intervals
        starts = df.start
        stops = df.end
        if initial_datetime is None:
            initial_datetime = starts.min()
        if final_datetime is None:
            final_datetime = stops.max()
        new_datetime = pd.date_range(initial_datetime, final_datetime, freq=freq)
        new_values = np.zeros(len(new_datetime))
        for idx, datetime in enumerate(new_datetime):
            next_datetime = datetime + pd.to_timedelta(freq)
            mask = (starts <= next_datetime) & (stops > datetime)
            if len(values[mask]) > 0:
                # calculate the fraction of each interval contained in the resampled interval
                value_fraction = interval_fraction(starts[mask], stops[mask], datetime, next_datetime)
                new_values[idx] = (values[mask] * value_fraction).agg(agg_method)
    else:
        # data is specified per datetime
        data_datetimes = df.datetime
        if initial_datetime is None:
            initial_datetime = data_datetimes.min()
        if final_datetime is None:
            final_datetime = data_datetimes.max()
        new_datetime = pd.date_range(initial_datetime, final_datetime, freq=freq)
        new_values = np.zeros(len(new_datetime))
        for idx, datetime in enumerate(new_datetime):
            next_datetime = datetime + pd.to_timedelta(freq)
            mask = (data_datetimes <= next_datetime) & (data_datetimes >= datetime)
            if len(values[mask]) > 0:
                new_values[idx] = values[mask].agg(agg_method)

    return pd.DataFrame({'datetime': new_datetime, name: new_values})

# %% ../nbs/api/05_readers.ipynb 17
WEARABLE_RESAMPLE_METHOD = {
    'steps': 'sum',
    'wake': 'max',
    'heartrate': 'mean',
    'light_estimate': 'mean',
    'activity': 'mean',
}

# %% ../nbs/api/05_readers.ipynb 18
def combine_wearable_dataframes(df_dict: Dict[str, pd.DataFrame], # dictionary of wearable dataframes 
                                resample_freq: str, # resampling frequency (e.g. '10min' for 10 minutes, see Pandas Offset aliases: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
                                metadata: Dict[str, str] = None, # metadata for the combined dataframe
                                ) -> pd.DataFrame: # combined wearable dataframe
    "Combine a dictionary of wearable dataframes into a single dataframe with resampling"
    df_list = []
    # find common initial and final datetimes
    initial_datetimes = []
    final_datetimes = []
    for name in df_dict.keys():
        df = df_dict[name]
        df.wearable.is_valid()
        if 'start' in df.columns:
            initial_datetimes.append(df.start.min())
            final_datetimes.append(df.end.max())
        else:
            initial_datetimes.append(df.datetime.min())
            final_datetimes.append(df.datetime.max())
    initial_datetime = min(initial_datetimes)
    final_datetime = max(final_datetimes)
    # resample each df
    for name in df_dict.keys():
        df = df_dict[name]
        new_df = resample_df(df, name, resample_freq, 
                             WEARABLE_RESAMPLE_METHOD[name],
                             initial_datetime=initial_datetime,
                             final_datetime=final_datetime)
        df_list.append(new_df)
    # merge all dfs by datetime
    df = df_list[0]
    for i in range(1, len(df_list)):
        df = df.merge(df_list[i], on='datetime', how='outer')
    # sort by datetime
    df.sort_values(by='datetime', inplace=True)
    # add metadata
    if metadata is not None:
        df.wearable.add_metadata(metadata, inplace=True)
    else:
        df.wearable.add_metadata({'data_id': 'combined_dataframe'}, inplace=True)
    return df



"""Useful functions for processing and analyzing circadian data"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/07_utils.ipynb.

# %% auto 0
__all__ = ['phase_difference', 'amplitude_percent_change', 'sleep_midpoint_and_duration', 'utc_to_hrs', 'phase_ic_guess',
           'abs_hour_diff', 'cut_phases_12', 'convert_binary', 'cal_days_diff', 'parse_dt', 'phase_coherence',
           'phase_coherence_clock', 'angle_difference', 'subtract_clock_times', 'times_to_angle']

# %% ../nbs/api/07_utils.ipynb 4
import scipy as sp
import numpy as np
import pandas as pd
from typing import List
from scipy.stats import circmean

# %% ../nbs/api/07_utils.ipynb 5
def phase_difference(phase_1: float, # phase between [-pi, pi]
                     phase_2: float # phase between [-pi, pi]
                     ) -> float:
    "Phase difference (phase_2 - phase_1). Negative values represent clockwise rotations"
    diff = phase_2 - phase_1
    normalized_diff = (diff + np.pi) % (2 * np.pi) - np.pi
    return normalized_diff

# %% ../nbs/api/07_utils.ipynb 7
def amplitude_percent_change(amplitude_1: float, # amplitude between (0, inf)
                             amplitude_2: float # amplitude between (0, inf)
                             ):
    "Percent change between two amplitudes"
    return (amplitude_2 - amplitude_1) / amplitude_1 * 100

# %% ../nbs/api/07_utils.ipynb 9
def sleep_midpoint_and_duration(
    time: np.ndarray, # array of time values
    sleep_state: np.ndarray, # array of sleep state values
) -> List[np.ndarray]:
    "Calculate sleep duration and mid-sleep time"
    if not isinstance(time, np.ndarray):
        time = np.array(time)
        raise ValueError("time must be a numpy array")
    if not isinstance(sleep_state, np.ndarray):
        sleep_state = np.array(sleep_state)
        raise ValueError("sleep_state must be a numpy array")
    if len(time) != len(sleep_state):
        raise ValueError("time and sleep_state must have the same length")

    sleep_start_idxs = np.where(np.diff(sleep_state) == 1)[0]
    sleep_end_idxs = np.where(np.diff(sleep_state) == -1)[0] 
    # trim any incomplete sleep windows
    if sleep_start_idxs[0] > sleep_end_idxs[0]:
        sleep_end_idxs = sleep_end_idxs[1:]
    if sleep_start_idxs[-1] > sleep_end_idxs[-1]:
        sleep_start_idxs = sleep_start_idxs[:-1]

    sleep_duration = np.mean(time[sleep_end_idxs] - time[sleep_start_idxs])
    sleep_midpoints = (time[sleep_start_idxs] + time[sleep_end_idxs]) / 2.0
    mid_sleep_time = circmean(np.mod(sleep_midpoints, 24.0), high=24.0)

    return mid_sleep_time, sleep_duration

# %% ../nbs/api/07_utils.ipynb 11
def utc_to_hrs(d: pd.Timestamp # UTC timestamp to convert
               ) -> float: # hours since midnight
    "Convert UTC timestamp to hours since midnight"
    return d.hour + d.minute / 60.0 + d.second / 3600.0

# %% ../nbs/api/07_utils.ipynb 13
def phase_ic_guess(time_of_day: float) -> float:
    time_of_day = np.fmod(time_of_day, 24.0)

    # Wake at 8 am after 8 hours of sleep
    # State at 00:00
    psi = 1.65238233

    # Convert to radians, add to phase
    psi += time_of_day * np.pi / 12
    return psi


def abs_hour_diff(x, y):
    "Find the difference in hours between two clock times (wrapped)"
    a1 = min(x, y)
    a2 = max(x, y)
    s1 = a2-a1
    s2 = 24.0+a1-a2
    return(min(s1, s2))


def cut_phases_12(p):
    """
    Function to make the branch cut for the DLMO times be at 12 instead of 24.
    This is better because lots of DLMOs are near midnight, but many fewer are near
    noon.

        cut_phases_12(timept)
    """

    while (p < 0.0):
        p += 24.0

    p = np.fmod(p, 24.0)

    if p > 12:
        return p-24.0
    else:
        return p


def convert_binary(x, breakpoint: float = 0.50):
    x[x <= breakpoint] = 0.0
    x[x > breakpoint] = 1.1
    return x


def cal_days_diff(a, b):
    "Get the calander days between two time dates"
    A = a.replace(hour=0, minute=0, second=0, microsecond=0)
    B = b.replace(hour=0, minute=0, second=0, microsecond=0)
    return (A - B).days


def parse_dt(date, time):
    strDate = date + ' ' + time
    return pd.to_datetime(strDate, format='%m/%d/%Y %I:%M %p')

# %% ../nbs/api/07_utils.ipynb 14
def phase_coherence(series):
    Z=complex(0,0)
    series=np.array(series)
    for i in range(len(series)):
        Z+=np.exp(series[i]*complex(0,1))

    Z=Z/float(len(series))

    ans=np.absolute(Z)
    return(ans)


def phase_coherence_clock(series):
    angles=np.pi/12.0*series
    return(phase_coherence(angles))


def angle_difference(c1, c2) -> float:
    "Find the angle between two angles given in radians"
    return(np.angle(np.exp(complex(0,1)*(c1-c2))))


def subtract_clock_times(c1, c2):
    "Find the hour differences between two clock times new"
    a1=sp.pi/12.0*c1
    a2=sp.pi/12.0*c2
    adiff=angle_difference(a1, a2)
    return(12.0/sp.pi*adiff)

# %% ../nbs/api/07_utils.ipynb 15
def times_to_angle(time_vector: np.ndarray):
    "Take an array of times and return R, psi giving the mean angle (psi) and amplitude (R)"
    rad_vector = np.fmod(time_vector, 24.0) * np.pi/12.0
    Z = np.sum(np.exp(rad_vector*1j))/len(rad_vector)
    return(np.abs(Z), np.angle(Z))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/09_cli.ipynb.

# %% auto 0
__all__ = ['mytz', 'main_acto', 'main_esri']

# %% ../nbs/api/09_cli.ipynb 4
import sys
import torch
import argparse
import circadian
import numpy as np
import pandas as pd
from torch import jit
from pathlib import Path
from pytz import timezone
from datetime import datetime
from .metrics import *
import matplotlib.pyplot as plt
from .plots import Actogram
from .models import Hannay19
from .utils import phase_ic_guess
from .utils import simple_norm_stepshr_sleep_classifier
from .sleep import cluster_sleep_periods_scipy, sleep_midpoint
from .readers import load_json, load_csv

mytz = timezone('EST')

# %% ../nbs/api/09_cli.ipynb 6
def main_acto():
    parser = argparse.ArgumentParser(description="""Make an actogram""")

    parser.add_argument('-j', '--json',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a json file with the data"
                        )
    parser.add_argument('-c', '--csv',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a directory with csv data in it"
                        )
    parser.add_argument("-sm", "--sleepmodel",
                        required=False,
                        action='store',
                        type=str,
                        help="Machine learning pytorch model for steps and heartrate to wake prediction"
                        )
    parser.add_argument('-d', '--dlmo',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Integrate the model and plot dlmo times"
                        )
    parser.add_argument('-cbt', '--cbt',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Integrate the model and plot core body temperature times"
                        )
    parser.add_argument('--sleep',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Add sleep midpoints"
                        )
    parser.add_argument('-t', '--threshold',
                        required=False,
                        action='store',
                        type=float,
                        help="Threshold for displaying as light on the actogram",
                        default=1.0
                        )
    parser.add_argument('-m', '--multiplier',
                        required=False,
                        action='store',
                        type=float,
                        help="Steps to light multiplier",
                        default=1.0
                        )
    parser.add_argument('-p', '--period',
                        required=False,
                        action='store',
                        default=23.84,
                        type=float,
                        help="Set the SPM period"
                        )
    parser.add_argument('-r', '--raw',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Plot the raw steps and heartrate data"
                        )
    parser.add_argument('--scatter',
                        required=False,
                        action='store_true',
                        default=False,
                        help="Plot the raw steps and heartrate data"
                        )
    parser.add_argument('-t1', '--t1',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data before this time, in days"
                        )
    parser.add_argument('-t2', '--t2',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data after this time, in days"
                        )
    parser.add_argument('-s', '--sigma',
                        required=False,
                        type=float,
                        action='store',
                        default=0.5,
                        help="Smooth the light(steps) data"
                        )


    args = parser.parse_args()

    def generated_sleep(ml_model_path, steps, hr):
        ml_model = jit.load(ml_model_path)
        data = torch.vstack((torch.tensor(steps), torch.tensor(hr)))
        data = simple_norm_stepshr_sleep_classifier(data).unsqueeze(0).float()
        wake_predicted = torch.sigmoid(ml_model(data)).squeeze(
            0).squeeze(0).detach().numpy()
        len_out = wake_predicted.shape[0]
        wake_predicted = np.hstack((wake_predicted, len(steps) - len_out))
        print(wake_predicted)
        wake_predicted = np.where(wake_predicted > 0.50, 1.0, 0.0)
        return wake_predicted
    
    if args.json:
        awObj = read_standard_json(args.json)

    if args.csv:
        awObj = read_standard_csv(args.csv)

    if args.csv or args.json:
        if args.t1 or args.t2:
            t1 = args.t1 or 0
            t2 = args.t2 or np.floor(np.squeeze(awObj.time_total)[-1]/24.0)+1
            awObj.trim_by_hour(t1*24.0, t2*24.0, inplace=True)
        hr = awObj.heartrate
        ts = awObj.time_total
        steps = awObj.steps

        if args.sleepmodel:
            pred_sleep = generated_sleep(
                ml_model_path=args.sleepmodel, 
                steps=steps, 
                hr=hr)
            print(pred_sleep)
            awObj._dataframe['wake'][0:len(pred_sleep)] = pred_sleep

    if args.raw:
        awObj.plot()

    ic = np.array([0.70, phase_ic_guess(ts[0]), 0.0])

    if args.scatter:
        awObj.scatter_hr_steps()


    plt.figure()
    ax = plt.gca()
    acto = Actogram(np.hstack(ts),
                    np.hstack(steps),
                    ax=ax,
                    threshold=args.threshold,
                    opacity=1.0,
                    sigma=[args.sigma, args.sigma])

    if awObj.wake is not None and args.sleep:
        print("Adding sleep layer")
        acto = Actogram(np.hstack(ts),
                        np.hstack(awObj.wake),
                        ax=ax,
                        threshold=0.5,
                        smooth=False,
                        opacity=0.50,
                        color='green')

    if args.dlmo:
        ic = np.array([0.70, phase_ic_guess(ts[0]), 0.0])
        hmodel = Hannay19({'tau': args.period})
        traj = hmodel(ts, args.multiplier*steps, ic)
        dlmos = hmodel.dlmos(traj)
        acto.plot_phasemarker(
            dlmos, error=np.ones(len(dlmos)), color='blue')
        print(f"DLMO mean: {np.mean(np.fmod(dlmos, 24.0))}")
        print("Last 14 days of DLMOs")
        print(np.fmod(dlmos[-14:], 24.0))

    if args.cbt:
        ic = np.array([0.70, phase_ic_guess(ts[0]), 0.0])
        hmodel = Hannay19({'tau': args.period})
        traj = hmodel(ts, args.multiplier*steps, ic)
        cbts = hmodel.cbt(traj)
        acto.plot_phasemarker(cbts, error=np.ones(len(cbts)), color='red')
        print(f"CBT mean: {np.mean(np.fmod(cbts, 24.0))}")
        print("Last 14 days of CBTs")
        print(np.fmod(cbts[-14:], 24.0))


    # if args.sleep:
    #     ts_flat = np.hstack(ts)
    #     steps_flat = np.hstack(steps)
    #     if awObj.wake is None:
    #         spm2 = Hannay19({'tau' : args.period})
    #         sol= spm2.integrate_model(ts_flat, args.multiplier*steps_flat, phase_ic_guess(ts_flat[0]))
    #         wake_score = np.diff(sol[3,:], prepend=0) > 0
    #     else:
    #         print("Using the provided wake column")
    #         wake_score=np.hstack(awObj.wake)

    #     sleep_clusters = cluster_sleep_periods_scipy(wake_score,
    #                                            90.0,
    #                                            makeplot=False)
    #     sleep_mid, duration = sleep_midpoint(ts_flat, sleep_clusters)
    #     acto.plot_phasemarker(sleep_mid, error=(
    #         duration/2.0), color="green", alpha=1.0)
    #     print(f"The median sleep duration is {np.median(duration)}")
    #     print(f"The average bedtime is {np.mean(np.fmod(sleep_mid-0.5*duration, 24.0))}")
    #     print(f"The average waketime is {np.mean(np.fmod(sleep_mid,24.0)+0.5*duration)}")

    plt.show()

# %% ../nbs/api/09_cli.ipynb 7
def main_esri():
    mytz = timezone('EST')
    parser = argparse.ArgumentParser(
        description="""Compute the Entrainment Signal Regularity Index for a data set""")

    parser.add_argument("-a", "--actogram",
                        required=False,
                        action='store_true',
                        default=False
                        )

    parser.add_argument('-j', '--json',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a json file with the data"
                        )
    parser.add_argument('-c', '--csv',
                        required=False,
                        action='store',
                        type=str,
                        help="Specify a directory with csv data in it"
                        )

    parser.add_argument('-s', '--sigma',
                        required=False,
                        type=float,
                        action='store',
                        default=0.5,
                        help="Smooth the light(steps) data"
                        )

    parser.add_argument('-t1', '--t1',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data before this time, in days"
                        )
    parser.add_argument('-t2', '--t2',
                        required=False,
                        type=float,
                        action='store',
                        default=None,
                        help="Trim data after this time, in days"
                        )

    parser.add_argument("-sm", "--sleepmodel",
                        required=False,
                        action='store',
                        type=str,
                        help="Machine learning pytorch model for steps and heartrate to wake prediction"
                        )

    parser.add_argument('-t', '--threshold',
                        required=False,
                        action='store',
                        type=float,
                        help="Threshold for displaying as light on the actogram",
                        default=1.0
                        )

    parser.add_argument('-o', '--outfile',
                        required=False,
                        action='store',
                        type=str,
                        help="Path to the outfile",
                        default="esri_results.csv"
                        )
    parser.add_argument('-np', '--noplot',
                        required=False,
                        action='store_true',
                        help="Use this to suppress the plot",
                        default=False
                        )


    args = parser.parse_args()

    def generated_sleep(ml_model_path, steps, hr):
        ml_model = jit.load(ml_model_path)
        data = torch.vstack((torch.tensor(steps), torch.tensor(hr)))
        data = simple_norm_stepshr_sleep_classifier(data).unsqueeze(0).float()
        wake_predicted = torch.sigmoid(ml_model(data)).squeeze(
            0).squeeze(0).detach().numpy()
        len_out = wake_predicted.shape[0]
        wake_predicted = np.hstack((wake_predicted, len(steps) - len_out))
        wake_predicted = np.where(wake_predicted > 0.50, 1.0, 0.0)
        return wake_predicted


    if args.json:
        awObj = load_json(args.json)

    if args.csv:
        awObj = load_csv(args.csv)

    if args.csv or args.json:
        if args.t1 or args.t2:
            t1 = args.t1 or 0
            t2 = args.t2 or np.floor(np.squeeze(awObj.time_total)[-1]/24.0)+1
            awObj.trim_data(t1*24.0, t2*24.0)
        hr = awObj.heartrate
        ts = awObj.time_total
        steps = awObj.steps

        if args.sleepmodel:
            awObj.wake = generated_sleep(
                ml_model_path=args.sleepmodel, steps=steps, hr=hr)


    def moving_average(x, w):
        return np.convolve(x, np.ones(w), 'valid') / w


    smooth_length = 100
    num_days_compactness = 4.5

    ts_compact, timestamps, compactness = esri_trajectory(
        awObj, gamma=0.0, multiplier=1.0, num_days=num_days_compactness)
    compactness_smooth = moving_average(compactness, smooth_length)
    plt.figure()
    ax1 = plt.gca()
    ax1.plot(np.array(compactness)*24.0, np.array(ts_compact) /
            24.0 + num_days_compactness,  alpha=0.50)
    ax1.plot(compactness_smooth*24.0,
            np.array(ts_compact[smooth_length-1:])/24.0 + num_days_compactness, color='red')
    ax1.set_title(f"ESRI")
    ax1.set_ylabel("Days")
    ax1.set_xlabel("ESRI Score")
    ax1.set_ylim((0, max(np.array(ts_compact) / 24.0)))
    ax1.invert_yaxis()



    print(f"Writing results to {args.outfile}")
    df = pd.DataFrame({'timestamp': timestamps, 'esri': compactness})
    df.to_csv(args.outfile, index=False)

    if not args.noplot:
        acto = Actogram(np.hstack(ts),
                        np.hstack(steps),
                        ax=ax1,
                        threshold=args.threshold,
                        opacity=1.0,
                        sigma=[args.sigma, args.sigma])

        plt.show()



"""Defines the main models, tools, and base classes for simulating and implementing mathematical models of circadian rhythms."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/00_models.ipynb.

# %% auto 0
__all__ = ['DynamicalTrajectory', 'CircadianModel', 'Forger99', 'Hannay19', 'Hannay19TP', 'Jewett99', 'Hilaire07', 'Breslow13',
           'Skeldon23']

# %% ../nbs/api/00_models.ipynb 4
import warnings
import numpy as np
from abc import ABC
from typing import Tuple
from scipy.signal import find_peaks
from fastcore.basics import patch_to
from .lights import LightSchedule

# %% ../nbs/api/00_models.ipynb 7
def _time_input_checking(time):
    "Checks if time is a valid input"
    if not isinstance(time, np.ndarray):
        raise TypeError("time must be a numpy array")
    if time.ndim != 1:
        raise ValueError("time must be a 1D array")
    if not np.issubdtype(time.dtype, np.number):
        raise TypeError("time must be numeric")
    if not np.all(np.diff(time) > 0):
        raise ValueError("time must be monotonically increasing")
    return True

def _state_input_checking(state, time):
    "Checks if state is a valid input"
    if not isinstance(state, np.ndarray):
        raise TypeError("states must be a numpy array")
    if state.ndim < 1:
        raise ValueError("states must have at least 1 dimension")
    if state.ndim > 3:
        raise ValueError("states must have at most 3 dimensions")
    if state.shape[0] != len(time):
        raise ValueError("states must have the same length as time")
    if not np.issubdtype(state.dtype, np.number):
        raise TypeError("states must be numeric")
    return True

# %% ../nbs/api/00_models.ipynb 9
class DynamicalTrajectory:
    "A class to store solutions of differential equation models that contains both the time points and the states"
    def __init__(self, 
                 time: np.ndarray, # time points
                 states: np.ndarray # state at time points
                 ) -> None:
        # input checking
        _time_input_checking(time)
        _state_input_checking(states, time)
        
        self.time = time
        self.states = states
        self.num_states = states.shape[1]
        if states.ndim >= 3:
            self.batch_size = states.shape[2]
        else:
            self.batch_size = 1

# %% ../nbs/api/00_models.ipynb 10
@patch_to(DynamicalTrajectory)
def __call__(self, timepoint: float) -> np.ndarray: # state of the system
    "Return the state at time t, linearly interpolated"
    # timepoint input checking
    if not isinstance(timepoint, (int, float)):
        raise TypeError("timepoint must be int or float")
    if timepoint < self.time[0] or timepoint > self.time[-1]:
        raise ValueError("timepoint must be within the time range")
    
    if self.batch_size == 1:
        values = np.zeros(self.num_states)
        for idx in range(self.num_states):
            values[idx] = np.interp(timepoint, self.time, self.states[..., idx])
    else:
        values = np.zeros((self.num_states, self.batch_size))
        for idx in range(self.num_states):
            for batch_idx in range(self.batch_size):
                values[idx, batch_idx] = np.interp(timepoint, self.time, self.states[..., idx, batch_idx])
    
    return values

# %% ../nbs/api/00_models.ipynb 11
@patch_to(DynamicalTrajectory)
def __getitem__(self, time_idx: int) -> Tuple[float, np.ndarray]:
    "Return the time and state at index idx"
    # idx input checking
    if not isinstance(time_idx, int):
        raise TypeError("idx must be int")
    if time_idx < -1 or time_idx >= len(self.time):
        raise ValueError(f"idx must be within 0 and {len(self.time)-1}, got {time_idx}")
    
    return self.time[time_idx], self.states[time_idx, ...]

# %% ../nbs/api/00_models.ipynb 12
@patch_to(DynamicalTrajectory)
def __len__(self) -> int:
    return len(self.time)

# %% ../nbs/api/00_models.ipynb 13
@patch_to(DynamicalTrajectory)
def get_batch(self, batch_idx: int) -> 'DynamicalTrajectory':
    "Obtain the trajectory for a single batch"
    # batch_idx input checking
    if not isinstance(batch_idx, int):
        raise TypeError("batch_idx must be int")
    if batch_idx < -1 or batch_idx >= self.batch_size:
        raise ValueError(f"batch_idx must be within -1 and {self.batch_size-1}, got {batch_idx}")
    if self.states.ndim >= 3:
        return DynamicalTrajectory(self.time, self.states[:, :, batch_idx])
    else:
        # no batch dimension
        return DynamicalTrajectory(self.time, self.states)

# %% ../nbs/api/00_models.ipynb 14
@patch_to(DynamicalTrajectory)
# String method
def __str__(self) -> str:
    time_str = np.array2string(self.time, precision=2, separator=", ", threshold=20)
    states_str = np.array2string(self.states, precision=2, separator=", ", threshold=20)
    output = f"Time:\n{time_str}\n"
    output += f"States:\n{states_str}"
    return output

# %% ../nbs/api/00_models.ipynb 17
def _parameter_input_checking(parameters):
    "Checks if parameters is a valid input for a circadian model"
    if not isinstance(parameters, dict):
        raise TypeError("parameters must be a dictionary")
    if len(parameters) == 0:
        raise ValueError("parameters must not be empty")
    for key, value in parameters.items():
        if not isinstance(key, str):
            raise TypeError("keys of parameters must be strings")
        if not isinstance(value, (int, float, type(None))):
            raise TypeError("values of parameters must be numeric or None")
    return True


def _positive_int_checking(number, name):
    "Checks if number is a positive integer"
    if not isinstance(number, int):
        raise TypeError(f"{name} must be an integer")
    if number < 1:
        raise ValueError(f"{name} must be positive")
    return True


def _initial_condition_input_checking(initial_condition, num_states):
    "Checks if initial_condition is a valid input for a circadian model"
    if not isinstance(initial_condition, np.ndarray):
        raise TypeError("initial_condition must be a numpy array")
    if not np.issubdtype(initial_condition.dtype, np.number):
        raise TypeError("initial_condition must be numeric")
    if initial_condition.shape[0] != num_states:
        raise ValueError(f"initial_condition must have length {num_states}")
    if np.any(np.isnan(initial_condition)):
        raise ValueError("initial_condition must not contain NaNs")
    return True


def _model_input_checking(input, num_inputs, time):
    "Checks if input has the correct shape and values for a circadian model"
    if not isinstance(input, np.ndarray):
        raise TypeError("input must be a numpy array")
    if not np.issubdtype(input.dtype, np.number):
        raise TypeError("input must be numeric")
    if input.shape[0] != len(time):
        raise ValueError(f"input's first dimension must have length {len(time)} based on the time array provided")
    if num_inputs > 1:
        if input.shape[1] != num_inputs:
            raise ValueError(f"input must have {num_inputs} columns")
    if np.any(np.isnan(input)):
        raise ValueError("input must not contain NaNs")

    
def _light_input_checking(light):
    "Checks if light is a valid input for a circadian model"
    if not isinstance(light, np.ndarray):
        raise TypeError("light must be a numpy array")
    if light.ndim != 1:
        raise ValueError("light must be a 1D array")
    if not np.issubdtype(light.dtype, np.number):
        raise TypeError("light must be numeric")
    if np.any(np.isnan(light)):
        raise ValueError("light must not contain NaNs")
    if not np.all(light >= 0):
        raise ValueError("light intensity must be nonnegative")
    return True


def _wake_input_checking(wake):
    "Checks if wake is a valid input for a circadian model"
    if not isinstance(wake, np.ndarray):
        raise TypeError("wake must be a numpy array")
    if wake.ndim != 1:
        raise ValueError("wake must be a 1D array")
    if not np.issubdtype(wake.dtype, np.number):
        raise TypeError("wake must be numeric")
    if np.any(np.isnan(wake)):
        raise ValueError("wake must not contain NaNs")
    if not np.all(wake >= 0) and not np.all(wake <= 1):
        raise ValueError("wake must be between 0 and 1")
    return True

# %% ../nbs/api/00_models.ipynb 19
class CircadianModel(ABC):
    "Abstract base class for circadian models that defines the common interface for all implementations"
    def __init__(self, 
                 default_params: dict, # default parameters for the model
                 num_states: int, # number of independent variables in the model
                 num_inputs: int, # number of inputs to the model such as light or wake state
                 default_initial_condition: np.ndarray # default initial conditions for the model
                 ) -> None:
        "Creates a new instance of the model"
        _parameter_input_checking(default_params)
        _positive_int_checking(num_states, "num_states")
        _positive_int_checking(num_inputs, "num_inputs")
        _initial_condition_input_checking(default_initial_condition, num_states)
        self.__default_params = default_params
        self._parameters = default_params
        for param_name, param_val in default_params.items():
            setattr(self, param_name, param_val)
        self.__num_states = num_states
        self.__num_inputs = num_inputs
        self.__default_initial_condition = default_initial_condition
        self._trajectory = None
        self._initial_condition = default_initial_condition
        self._min_marker_distance_in_hours = 13.0 # minimum allowed distance between CBTmin values


    @property
    def parameters(self) -> dict:
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        for param_name, param_val in value.items():
            setattr(self, param_name, param_val)
        self._parameters = value

    @property
    def _default_params(self) -> dict:
        return self.__default_params

    @_default_params.setter
    def _default_params(self, value):
        self._parameters = value
        self.__default_params = value

    @property
    def _num_states(self) -> int: # number of independent variables in the model
        return self.__num_states

    @_num_states.setter
    def _num_states(self, value):
        self.__num_states = value

    @property
    def _num_inputs(self) -> int: # number of inputs to the model such as light or wake state
        return self.__num_inputs
    
    @_num_inputs.setter
    def _num_inputs(self, value):
        self.__num_inputs = value

    @property
    def _default_initial_condition(self) -> np.ndarray:
        return self.__default_initial_condition
    
    @_default_initial_condition.setter
    def _default_initial_condition(self, value):
        self.__default_initial_condition = value

    @property
    def trajectory(self) -> DynamicalTrajectory:
        return self._trajectory
    
    @trajectory.setter
    def trajectory(self, value):
        self._trajectory = value

    @property
    def initial_condition(self) -> np.ndarray:
        return self._initial_condition
    
    @initial_condition.setter
    def initial_condition(self, value):
        self._initial_condition = value

# %% ../nbs/api/00_models.ipynb 20
@patch_to(CircadianModel)
def derv(self,
         t: float, # time
         state: np.ndarray, # dynamical state of the model
         input: np.ndarray, # inputs to the model such as light or wake state
         ) -> np.ndarray:
    "Right-hand-side of the differential equation model with state and light as inputs"
    return NotImplementedError("derv is not implemented for this model")

# %% ../nbs/api/00_models.ipynb 21
@patch_to(CircadianModel)
def step_rk4(self,
             t: float, # time
             state: np.ndarray, # dynamical state of the model
             input: np.ndarray, # inputs to the model such as light or wake state
             dt: float, # step size in hours 
             ) -> np.ndarray:
    "Integrate the state of the model for one timestep using a fourth-order Runge-Kutta algorithm. Assumes a constant light value for the time step"
    k1 = self.derv(t, state, input)
    k2 = self.derv(t, state + k1 * dt / 2.0, input)
    k3 = self.derv(t, state + k2 * dt / 2.0, input)
    k4 = self.derv(t, state + k3 * dt, input)
    state = state + (dt / 6.0) * (k1 + 2.0*k2 + 2.0*k3 + k4)
    return state

# %% ../nbs/api/00_models.ipynb 22
@patch_to(CircadianModel)
def integrate(self,
              time: np.ndarray, # time points for integration. Time difference between consecutive values determines step size of the solver
              initial_condition: np.ndarray=None, # initial state of the model
              input: np.ndarray=None, # model input (such as light or wake) for each time point 
              ) -> DynamicalTrajectory:
    "Solve the model for specific timepoints given initial conditions and model inputs"
    # input checking
    _time_input_checking(time)
    if input is None:
        raise ValueError("a model input must be provided via the input argument")
    else:
        _model_input_checking(input, self._num_inputs, time)
    if initial_condition is None:
        initial_condition = self._default_initial_condition
    else:
        _initial_condition_input_checking(initial_condition, self._num_states)
    
    self.initial_condition = initial_condition
    
    n = len(time)
    sol = np.zeros((n, *initial_condition.shape))
    sol[0,...] = initial_condition
    state = initial_condition

    for idx in range(1, n):
        t = time[idx]
        dt = t - time[idx-1]
        input_value = input[idx,...]
        state = self.step_rk4(t, state, input_value, dt)
        sol[idx,...] = state
    
    self._trajectory = DynamicalTrajectory(time, sol)
    return self._trajectory

# %% ../nbs/api/00_models.ipynb 23
@patch_to(CircadianModel)
def __call__(self,
             time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
             initial_condition: np.ndarray=None, # initial state of the model
             input: np.ndarray=None, # model input (such as light or wake) for each time point
             ):
    "Wrapper to integrate"
    return self.integrate(time, initial_condition, input)

# %% ../nbs/api/00_models.ipynb 24
@patch_to(CircadianModel)
def get_parameters_array(self)-> np.array:
    "Returns the parameters for the model as a numpy array"
    parameter_array = np.zeros(len(self.parameters))
    for idx, value in enumerate(self.parameters.values()):
        parameter_array[idx] = value
    return parameter_array

# %% ../nbs/api/00_models.ipynb 25
@patch_to(CircadianModel)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase for. If None, the phase is calculated for the current trajectory 
          time: float=None # timepoint to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    "Calculates the phase of the model at a given timepoint"
    raise NotImplementedError("phase is not implemented for this model")

# %% ../nbs/api/00_models.ipynb 26
@patch_to(CircadianModel)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude for. If None, the amplitude is calculated for the current trajectory 
              time: float=None, # timepoint to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    "Calculates the amplitude of the model at a given timepoint"
    raise NotImplementedError("amplitude is not implemented for this model")

# %% ../nbs/api/00_models.ipynb 27
@patch_to(CircadianModel)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt for. If None, the cbt is calculated for the current trajectory
        ) -> np.ndarray: # array of times when the cbt occurs
    "Finds the core body temperature minumum markers along a trajectory"
    raise NotImplementedError("cbt is not implemented for this model")

# %% ../nbs/api/00_models.ipynb 28
@patch_to(CircadianModel)
def dlmos(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the dlmos for. If None, the dlmos are calculated for the current trajectory
          ) -> np.ndarray: # array of times when dlmo occurs 
    "Finds the Dim Light Melatonin Onset (DLMO) markers along a trajectory"
    raise NotImplementedError("dlmo is not implemented for this model")

# %% ../nbs/api/00_models.ipynb 29
@patch_to(CircadianModel)
def equilibrate(self,
                time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                input: np.ndarray, # model input (such as light or wake) for each time point
                num_loops: int=10 # number of times to loop the input
                ) -> np.ndarray: # final state of the model
    "Equilibrate the model by looping the given light_estimate. Assumes the schedule repeats periodically after it ends"
    # input checking
    _time_input_checking(time)
    _model_input_checking(input, self._num_inputs, time)
    _positive_int_checking(num_loops, "num_loops")
    
    initial_condition = self._default_initial_condition
    dlmos = []
    for _ in range(num_loops):
        sol = self.integrate(time, initial_condition, input).states
        dlmos.append(self.dlmos())
        initial_condition = sol[-1, ...]
    # check that the output is equilibrated
    is_equilibrated = np.all(np.isclose(dlmos[-1][-1], dlmos[-2][-1], atol=1e-3))
    if not is_equilibrated:
        warnings.warn("The model did not equilibrate. Try increasing the number of loops.")
    final_state = sol[-1, ...]
    return final_state

# %% ../nbs/api/00_models.ipynb 30
def _get_default_initial_condition(
        model: CircadianModel, # model to calculate the default initial condition for
        num_loops: int=10 # number of times to loop the regular schedule
        ) -> np.ndarray:
    "Calculates a default initial condition by simulating the given model on a 16 hour light, 8 hour darkness schedule until convergence"
    # input checking
    if not isinstance(model, CircadianModel):
        raise TypeError("model must be a CircadianModel")
    model._default_initial_condition = 0.5 * np.ones(model._num_states)
    schedule = LightSchedule.Regular(lights_on=8, lights_off=24)    
    time = np.arange(0.0, 24.0, 0.1)
    if model._num_inputs == 1:
        light_input = schedule(time)
        default_initial_condition = model.equilibrate(time, light_input, num_loops)
    elif model._num_inputs == 2:
        light_input = schedule(time)
        wake_input = np.zeros_like(light_input)
        wake_input[light_input > 0] = 0
        wake_input[light_input == 0] = 1
        input = np.stack((light_input, wake_input), axis=1)
        default_initial_condition = model.equilibrate(time, input, num_loops)
    return default_initial_condition

def _check_cbtmin_spacing(
        cbtmin_times: np.ndarray, # array of times when the cbtmin occurs
        min_spacing: float=6.0, # minimum spacing between cbtmin markers
        ) -> bool:
    "Checks if the spacing between cbtmin markers is valid"
    cbtmin_spacing = np.diff(cbtmin_times)
    if np.any(cbtmin_spacing < min_spacing):
        # raise a warning
        warnings.warn(f"The data contains cbtmin markers that are spaced by less than {min_spacing} hours. Removal of duplicate cbtmin markers is recommended.")

# %% ../nbs/api/00_models.ipynb 32
class Forger99(CircadianModel): 
    "Implementation of Forger's 1999 model from the article 'A simpler model of the human circadian pacemaker'"
    def __init__(self, params=None):
        default_params = {
            'taux': 24.2, 'mu': 0.23, 'G': 33.75, 
            'alpha_0': 0.05, 'beta': 0.0075, 'p': 0.50, 
            'I0': 9500.0, 'k': 0.55, 'cbt_to_dlmo': 7.0,
            }
        num_states = 3 # x, xc, n
        num_inputs = 1 # light
        default_initial_condition = np.array([-0.0843259, -1.09607546, 0.45584306]) # condition at midnight for 16L, 8D schedule
        super(Forger99, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params

    def integrate(self,
                  time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                  initial_condition: np.ndarray=None, # initial state of the model
                  input: np.ndarray=None, # model input (such as light or wake) for each time point
                  ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking for Forger99
        if input is not None:
            _light_input_checking(input)
        return super().integrate(time, initial_condition, input)

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Forger99"

# %% ../nbs/api/00_models.ipynb 33
@patch_to(Forger99)
def derv(self, 
         t: float, # time
         state: np.ndarray, # dynamical state (x, xc, n)
         input: float # light intensity in lux
         ) -> np.ndarray: # derivative of the state
     "Right-hand-side of the differential equation model"
     x = state[0,...]
     xc = state[1,...]
     n = state[2,...]
     light = input

     alpha = self.alpha_0 * pow((light / self.I0), self.p)
     Bhat = self.G * (1.0 - n) * alpha * (1 - 0.4 * x) * (1 - 0.4 * xc)
     mu_term = self.mu * (xc - 4.0 / 3.0 * pow(xc, 3.0))
     taux_term = pow(24.0 / (0.99669 * self.taux), 2.0) + self.k * Bhat

     dydt = np.zeros_like(state)
     dydt[0,...] = np.pi / 12.0 * (xc + Bhat)
     dydt[1,...] = np.pi / 12.0 * (mu_term - x * taux_term)
     dydt[2,...] = 60.0 * (alpha * (1.0 - n) - self.beta * n)

     return dydt

# %% ../nbs/api/00_models.ipynb 34
@patch_to(Forger99)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 35
@patch_to(Forger99)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.sqrt(x**2 + y**2)

# %% ../nbs/api/00_models.ipynb 36
@patch_to(Forger99)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the minimum of x"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    dt = np.diff(trajectory.time)[0]
    inverted_x = -1*trajectory.states[:,0]
    cbt_min_idxs, _ = find_peaks(inverted_x, distance=np.ceil(self._min_marker_distance_in_hours / dt)) # min separation between troughs
    cbtmin_times = trajectory.time[cbt_min_idxs]
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 37
@patch_to(Forger99)
def dlmos(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the dlmo. If None, the current trajectory is used 
          ) -> np.ndarray:
    "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    return self.cbt(trajectory) - self.cbt_to_dlmo

# %% ../nbs/api/00_models.ipynb 40
class Hannay19(CircadianModel):
    "Implementation of Hannay's 2019 single population model from the article 'Macroscopic models for human circadian rhythms'"
    def __init__(self, params=None):
        default_params = {
            'tau': 23.84, 'K': 0.06358, 'gamma': 0.024, 
            'Beta1': -0.09318, 'A1': 0.3855, 'A2': 0.1977, 
            'BetaL1': -0.0026, 'BetaL2': -0.957756, 'sigma': 0.0400692, 
            'G': 33.75, 'alpha_0': 0.05, 'delta': 0.0075, 'p': 1.5, 'I0': 9325.0,
            'cbt_to_dlmo': 7.0}
        num_states = 3 # R, Psi, n
        num_inputs = 1 # light
        default_initial_condition = np.array([0.82041911, 1.71383697, 0.52318122]) # condition at midnight for 16L, 8D schedule
        super(Hannay19, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params

    def integrate(self,
                time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                initial_condition: np.ndarray=None, # initial state of the model
                input: np.ndarray=None, # model input (such as light or wake) for each time point
                ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking for Hannay19
        if input is not None:
            _light_input_checking(input)
        return super().integrate(time, initial_condition, input)

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Hannay19"

# %% ../nbs/api/00_models.ipynb 41
@patch_to(Hannay19)
def derv(self,
         t: float, # time
         state: np.ndarray, # dynamical state (R, Psi, n)
         input: float # light intensity in lux
         ) -> np.ndarray: # derivative of the state
    "Right-hand-side of the differential equation model"
    R = state[0,...]
    Psi = state[1,...]
    n = state[2,...]
    light = input   

    alpha = self.alpha_0 * pow(light, self.p) / (pow(light, self.p) + self.I0)

    Bhat = self.G * (1.0 - n) * alpha
    A1_term_amp = self.A1 * 0.5 * Bhat * (1.0 - pow(R, 4.0)) * np.cos(Psi + self.BetaL1)
    A2_term_amp = self.A2 * 0.5 * Bhat * R * (1.0 - pow(R, 8.0)) * np.cos(2.0 * Psi + self.BetaL2)
    LightAmp = A1_term_amp + A2_term_amp
    A1_term_phase = self.A1 * Bhat * 0.5 * (pow(R, 3.0) + 1.0 / R) * np.sin(Psi + self.BetaL1)
    A2_term_phase = self.A2 * Bhat * 0.5 * (1.0 + pow(R, 8.0)) * np.sin(2.0 * Psi + self.BetaL2)
    LightPhase = self.sigma * Bhat - A1_term_phase - A2_term_phase

    dydt = np.zeros_like(state)
    dydt[0,...] = -1.0 * self.gamma * R + self.K * np.cos(self.Beta1) / 2.0 * R * (1.0 - pow(R, 4.0)) + LightAmp
    dydt[1,...] = 2*np.pi/self.tau + self.K / 2.0 * np.sin(self.Beta1) * (1 + pow(R, 4.0)) + LightPhase
    dydt[2,...] = 60.0 * (alpha * (1.0 - n) - self.delta * n)

    return dydt

# %% ../nbs/api/00_models.ipynb 42
@patch_to(Hannay19)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = np.cos(trajectory.states[:, 1])
        y = np.sin(trajectory.states[:, 1])
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = np.cos(state[1])
            y = np.sin(state[1])
    return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 43
@patch_to(Hannay19)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        amplitude = trajectory.states[:, 0]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            amplitude = state[0] 
    return amplitude

# %% ../nbs/api/00_models.ipynb 44
@patch_to(Hannay19)
def cbt(self,
        trajectory: DynamicalTrajectory=None # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the times where the phase is pi"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    dt = np.diff(trajectory.time)[0]
    inverted_x = -np.cos(trajectory.states[:,1])
    cbt_min_idxs, _ = find_peaks(inverted_x, distance=np.ceil(self._min_marker_distance_in_hours / dt)) # min separation between troughs
    cbtmin_times = trajectory.time[cbt_min_idxs]
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 45
@patch_to(Hannay19)
def dlmos(self,
          trajectory: DynamicalTrajectory=None # trajectory to calculate the dlmo. If None, the current trajectory is used
          ) -> np.ndarray:
    "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    return self.cbt(trajectory) - self.cbt_to_dlmo

# %% ../nbs/api/00_models.ipynb 48
class Hannay19TP(CircadianModel):
    "Implementation of Hannay's 2019 two population model from the article 'Macroscopic models for human circadian rhythms'"
    def __init__(self, params=None):
        default_params = {
            'tauV': 24.25, 'tauD': 24.0, 'Kvv': 0.05, 
            'Kdd': 0.04, 'Kvd': 0.05, 'Kdv': 0.01, 
            'gamma': 0.024, 'A1': 0.440068, 'A2': 0.159136, 
            'BetaL': 0.06452, 'BetaL2': -1.38935, 'sigma': 0.0477375, 
            'G': 33.75, 'alpha_0': 0.05, 'delta': 0.0075, 'p': 1.5, 
            'I0': 9325.0, 'cbt_to_dlmo': 7.0,
            }
        num_states = 5 # Rv, Rd, Psiv, Psid, n
        num_inputs = 1 # light
        default_initial_condition = np.array([0.82423745, 0.82304996, 1.75233424, 1.863457, 0.52318122]) # condition at midnight for 16L, 8D schedule
        super(Hannay19TP, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params

    def integrate(self,
                time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                initial_condition: np.ndarray=None, # initial state of the model
                input: np.ndarray=None, # model input (such as light or wake) for each time point
                ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking for Hannay19TP
        if input is not None:
            _light_input_checking(input)
        return super().integrate(time, initial_condition, input)

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Hannay19TP"

# %% ../nbs/api/00_models.ipynb 49
@patch_to(Hannay19TP)
def derv(self,
         t: float, # time
         state: np.ndarray, # dynamical state (Rv, Rd, Psiv, Psid, n)
         input: float # light intensity in lux
         ) -> np.ndarray: # derivative of the state
     "Right-hand-side of the differential equation model"
     Rv = state[0,...]
     Rd = state[1,...]
     Psiv = state[2,...]
     Psid = state[3,...]
     n = state[4,...]
     light = input

     alpha = self.alpha_0 * pow(light, self.p) / (pow(light, self.p) + self.I0)
     Bhat = self.G * (1.0 - n) * alpha

     A1_term_amp = self.A1 * 0.5 * Bhat * (1.0 - pow(Rv, 4.0)) * np.cos(Psiv + self.BetaL)
     A2_term_amp = self.A2 * 0.5 * Bhat * Rv * (1.0 - pow(Rv, 8.0)) * np.cos(2.0 * Psiv + self.BetaL2)
     LightAmp = A1_term_amp + A2_term_amp
     A1_term_phase = self.A1 * Bhat * 0.5 * (pow(Rv, 3.0) + 1.0 / Rv) * np.sin(Psiv + self.BetaL)
     A2_term_phase = self.A2 * Bhat * 0.5 * (1.0 + pow(Rv, 8.0)) * np.sin(2.0 * Psiv + self.BetaL2)
     LightPhase = self.sigma * Bhat - A1_term_phase - A2_term_phase

     dydt = np.zeros_like(state)
     dydt[0,...] = -self.gamma * Rv + self.Kvv / 2.0 * Rv * (1 - pow(Rv, 4.0)) + self.Kdv / 2.0 * Rd * (1 - pow(Rv, 4.0)) * np.cos(Psid - Psiv) + LightAmp
     dydt[1,...] = -self.gamma * Rd + self.Kdd / 2.0 * Rd * (1 - pow(Rd, 4.0)) + self.Kvd / 2.0 * Rv * (1.0 - pow(Rd, 4.0)) * np.cos(Psid - Psiv)
     dydt[2,...] = 2.0 * np.pi / self.tauV + self.Kdv / 2.0 * Rd * (pow(Rv, 3.0) + 1.0 / Rv) * np.sin(Psid - Psiv) + LightPhase
     dydt[3,...] = 2.0 * np.pi / self.tauD - self.Kvd / 2.0 * Rv * (pow(Rd, 3.0) + 1.0 / Rd) * np.sin(Psid - Psiv)
     dydt[4,...] = 60.0 * (alpha * (1.0 - n) - self.delta * n)

     return dydt

# %% ../nbs/api/00_models.ipynb 50
@patch_to(Hannay19TP)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        time = trajectory.time
        x = np.cos(trajectory.states[:, 2])
        y = np.sin(trajectory.states[:, 2])
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = np.cos(state[2])
            y = np.sin(state[2])
    return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 51
@patch_to(Hannay19TP)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        time = trajectory.time
        amplitude = trajectory.states[:, 0]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            amplitude = state[0] 
    return amplitude

# %% ../nbs/api/00_models.ipynb 52
@patch_to(Hannay19TP)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the times where the phase is pi"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    dt = np.diff(trajectory.time)[0]
    inverted_x = -np.cos(trajectory.states[:,2])
    cbt_min_idxs, _ = find_peaks(inverted_x, distance=np.ceil(self._min_marker_distance_in_hours / dt)) # min separation between troughs
    cbtmin_times = trajectory.time[cbt_min_idxs]
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 53
@patch_to(Hannay19TP)
def dlmos(self,
          trajectory: DynamicalTrajectory=None # trajectory to calculate the dlmo. If None, the current trajectory is used
          ) -> np.ndarray:
    "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    return self.cbt(trajectory) - self.cbt_to_dlmo

# %% ../nbs/api/00_models.ipynb 56
class Jewett99(CircadianModel):
    "Implementation of Jewett's 1999 model from the article 'Revised Limit Cycle Oscillator Model of Human Circadian Pacemaker'"
    def __init__(self, params=None):
        default_params = {
            'taux': 24.2, 'mu': 0.13, 'G': 19.875,
            'beta': 0.013, 'k': 0.55, 'q': 1.0/3,
            'I0': 9500, 'p': 0.6, 'alpha_0': 0.16,
            'phi_ref': 0.8, 'cbt_to_dlmo': 7.0}
        num_states = 3 # x, xc, n
        num_inputs = 1 # light
        default_initial_condition = np.array([-0.10097101, -1.21985662, 0.50529415]) # condition at midnight for 16L, 8D schedule
        super(Jewett99, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params

    def integrate(self,
                  time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                  initial_condition: np.ndarray=None, # initial state of the model
                  input: np.ndarray=None, # model input (such as light or wake) for each time point
                  ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking for Jewett99
        if input is not None:
            _light_input_checking(input)
        return super().integrate(time, initial_condition, input)

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Jewett99"

# %% ../nbs/api/00_models.ipynb 57
@patch_to(Jewett99)
def derv(self,
         t: float, # time
         state: np.ndarray, # dynamical state (x, xc, n)
         input: float # light intensity in lux
         ) -> np.ndarray: # derivative of the state
    "Right-hand-side of the differential equation model"
    x = state[0,...]
    xc = state[1,...]
    n = state[2,...]
    light = input

    alpha = self.alpha_0 * (light / self.I0) ** self.p
    Bhat = self.G * alpha * (1 - n) * (1 - 0.4 * x) * (1 - 0.4 * xc)
    mu_term = self.mu * (1.0/3.0 * x + 4.0/3.0 * x**3 - 256.0/105.0 * x**7)
    taux_term = pow(24.0 / (0.99729 * self.taux), 2) + self.k * Bhat 

    dydt = np.zeros_like(state)
    dydt[0,...] = np.pi/12 * (xc + mu_term + Bhat)
    dydt[1,...] = np.pi/12 * (self.q * Bhat * xc - x * taux_term)
    dydt[2,...] = 60.0 * (alpha * (1 - n) - self.beta * n)
    
    return dydt

# %% ../nbs/api/00_models.ipynb 58
@patch_to(Jewett99)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 59
@patch_to(Jewett99)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.sqrt(x**2 + y**2)

# %% ../nbs/api/00_models.ipynb 60
@patch_to(Jewett99)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the corrected minimum of x"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    dt = np.diff(trajectory.time)[0]
    inverted_x = -1*trajectory.states[:,0]
    cbt_min_idxs, _ = find_peaks(inverted_x, distance=np.ceil(self._min_marker_distance_in_hours / dt)) # min separation between troughs
    cbtmin_times = trajectory.time[cbt_min_idxs] + self.phi_ref
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 61
@patch_to(Jewett99)
def dlmos(self,
          trajectory: DynamicalTrajectory=None # trajectory to calculate the dlmo. If None, the current trajectory is used
          ) -> np.ndarray:
    "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    return self.cbt(trajectory) - self.cbt_to_dlmo

# %% ../nbs/api/00_models.ipynb 64
class Hilaire07(CircadianModel):
    "Implementation of Hilaire's 2007 model from the article 'Addition of a non-photic component to a light-based mathematical model of the human circadian pacemaker'"
    def __init__(self, params=None):
        default_params = {
            'taux': 24.2, 'G': 37.0, 'k': 0.55, 'mu': 0.13, 'beta': 0.007, 
            'q': 1.0/3.0, 'rho': 0.032, 'I0': 9500.0, 'p': 0.5, 'a0': 0.1, 
            'phi_xcx': -2.98, 'phi_ref': 0.97, 'cbt_to_dlmo': 7.0,
            }
        num_states = 3 # x, xc, n
        num_inputs = 2 # light, wake
        default_initial_condition = np.array([-0.0480751, -1.22504441, 0.51854818])
        super(Hilaire07, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params

    def integrate(self,
                  time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                  initial_condition: np.ndarray=None, # initial state of the model
                  input: np.ndarray=None, # model input (such as light or wake) for each time point
                  ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking
        if input is not None:
            _light_input_checking(input[0,...])
            _wake_input_checking(input[1,...])
        return super().integrate(time, initial_condition, input)

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Hilaire07"

# %% ../nbs/api/00_models.ipynb 65
@patch_to(Hilaire07)
def derv(self,
         t: float, # time
         state: np.ndarray, # dynamical state (x, xc, n)
         input: np.ndarray # model input (light, wake)
         ) -> np.ndarray: # derivative of the state
     "Right-hand-side of the differential equation model"
     x = state[0,...]
     xc = state[1,...]
     n = state[2,...]
     light = input[0,...] 
     wake = input[1,...]
     
     alpha = self.a0 * (np.power(light / self.I0, self.p)) * (light / (light + 100.0))
     Bhat = self.G * (1 - n) * alpha * (1 - 0.4 * x) * (1 - 0.4 * xc) 
     # From article: sigma equals either 1 (for sleep/rest) or 0 (for wake/activity),
     sigma = 1.0 if wake < 0.5 else 0.0
     # Calculate psi_cx
     C = t % 24
     CBTmin = self.phi_xcx + self.phi_ref
     CBTminlocal = CBTmin * 24.0 / (2*np.pi)
     psi_cx = C - CBTminlocal
     psi_cx = psi_cx % 24
     # Define Ns
     Nsh = self.rho * (1.0/3.0 - sigma)
     if (psi_cx > 16.5 and psi_cx < 21.0):
          Nsh = self.rho * (1.0/3.0)
     Ns = Nsh * (1 - np.tanh(10.0 * x))

     mu_term = self.mu * (1.0 / 3.0 * x + 4.0 / 3.0 * np.power(x, 3.0) - 256.0 / 105.0 * np.power(x, 7.0))
     taux_term = (np.power((24.0 / (0.99729 * self.taux)), 2) + self.k * Bhat)
     
     dydt = np.zeros_like(state)
     dydt[0,...] = np.pi / 12.0 * (xc + mu_term + Bhat + Ns)
     dydt[1,...] = np.pi / 12.0 * (self.q * Bhat * xc - x * taux_term)
     dydt[2,...] = 60.0 * (alpha * (1.0 - n) - self.beta * n)
     
     return dydt

# %% ../nbs/api/00_models.ipynb 66
@patch_to(Hilaire07)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 67
@patch_to(Hilaire07)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.sqrt(x**2 + y**2)

# %% ../nbs/api/00_models.ipynb 68
@patch_to(Hilaire07)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the corrected minimum of x"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    dt = np.diff(trajectory.time)[0]
    inverted_x = -1*trajectory.states[:,0]
    cbt_min_idxs, _ = find_peaks(inverted_x, distance=np.ceil(self._min_marker_distance_in_hours / dt)) # min separation between troughs
    cbtmin_times = trajectory.time[cbt_min_idxs] + self.phi_ref
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 69
@patch_to(Hilaire07)
def dlmos(self,
          trajectory: DynamicalTrajectory=None # trajectory to calculate the dlmo. If None, the current trajectory is used
          ) -> np.ndarray:
    "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    return self.cbt(trajectory) - self.cbt_to_dlmo

# %% ../nbs/api/00_models.ipynb 72
class Breslow13(CircadianModel):
    "Implementation of melatonin suppression model from Breslow's 2013 article 'A Mathematical Model of the Circadian Phase-Shifting Effects of Exogenous Melatonin'"
    def __init__(self, params=None):
        default_params = {
            'H1_threshold': 0.001, 
            'K': 0.55, # not specified in paper, taken from Jewett99 and Hilaire07
            'I0': 9500.0, 'I1': 100.0, 'alpha_0': 0.1, 'beta': 0.007,
            'p': 0.5, 'r': 15.36, 'G': 37.0, 'b': 0.4, 'gamma': 0.13,
            'kappa': 12.0/np.pi, 'tauc': 24.1, 'f': 0.99729, 'm': 7.0,
            'eta': 0.04, 'xi': 0.54, 
            # the following parameters have been converted from seconds to hours
            'betaIP': 0.000783 * 3600.0, 'betaCP': 0.000335 * 3600.0,
            'betaAP': 0.000162 * 3600.0, 'a': 0.0010422 * 3600.0, 
            # threshold phases have been confined to [-pi, pi]
            'phi_on': 6.113 - 2 * np.pi, 'phi_off': 4.352 - 2 * np.pi,
            'delta': 600.0, 'Mmax': 0.019513, 'Hsat': 861.0, 'sigma': 50.0, 
            'phi_ref': 0.97, 'cbt_to_dlmo': 7.0,
            }
        num_states = 6 # x, xc, n, H1, H2, H3
        num_inputs = 1 # light
        default_initial_condition = np.array([-0.219, -1.22, 0.519, 0.0, 0.0, 0.0])

        super(Breslow13, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params

    def integrate(self,
                  time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                  initial_condition: np.ndarray=None, # initial state of the model
                  input: np.ndarray=None, # model input (such as light or wake) for each time point
                  ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking
        if input is not None:
            _light_input_checking(input)
        return super().integrate(time, initial_condition, input)

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Breslow13"
    
    def avoid_negative_H1(self, H1, B):
        if H1 < self.H1_threshold and (1.0 - self.m * B) < 0:
            return 0.0
        else:
            return 1.0
    
    def melatonin_drive(self, H2):
        exp_factor = 1.0 + np.exp((self.Hsat - H2)/ self.sigma)
        return self.Mmax / exp_factor
    

# %% ../nbs/api/00_models.ipynb 73
@patch_to(Breslow13)
def derv(self,
         t: float, # time
         state: np.ndarray, # dynamical state (x, xc, n, H1, H2, H3)
         input: np.ndarray # model input (light, wake)
         ) -> np.ndarray: # derivative of the state
     "Right-hand-side of the differential equation model"
     x = state[0,...]
     xc = state[1,...]
     n = state[2,...]
     H1 = state[3,...] # pineal melatonin
     H2 = state[4,...] # plasma melatonin
     H3 = state[5,...] # exogenous melatonin
     light = input
     
     alpha = self.alpha_0 * ((light / self.I0) ** self.p) * (light / (light + self.I1))
     B = self.G * alpha * (1.0 - n) * (1.0 - self.b * x) * (1.0 - self.b * xc)
     gamma_term = self.gamma * (x / 3.0 + 4.0 / 3.0 * x**3 - 256.0 / 105.0 * x**7)
     tauc_term = pow(24.0 / (self.f * self.tauc), 2) + self.K * B 
     M = self.melatonin_drive(H2)
     S_H1_B = self.avoid_negative_H1(H1, B)

     phase = np.arctan2(x, xc) # note this is the opposite to all other models (here xc is on the x-axis)
     if phase < self.phi_on and phase > self.phi_off:
          activation = 1.0 - np.exp(-self.delta * np.mod(self.phi_on - phase, 2*np.pi))
          normalization = 1.0 - np.exp(-self.delta * np.mod(self.phi_on - self.phi_off, 2*np.pi))
          A = self.a * activation / normalization
     else:
          A = self.a * np.exp(-self.r * np.mod(self.phi_on - self.phi_off, 2*np.pi))
     
     dydt = np.zeros_like(state)
     dydt[0,...] = (1.0/self.kappa) * (xc + gamma_term + B - self.eta * M)
     dydt[1,...] = (1.0/self.kappa) * (B * xc / 3.0 - x * tauc_term - self.xi * M)
     dydt[2,...] = 60.0 * (alpha * (1 - n) - self.beta * n) # typo on paper missing 60.0
     dydt[3,...] = - self.betaIP * H1 + A * (1.0 - self.m * B) * S_H1_B
     dydt[4,...] = self.betaIP * H1 - self.betaCP * H2 + self.betaAP * H3
     dydt[5,...] = - self.betaAP * H3

     return dydt

# %% ../nbs/api/00_models.ipynb 74
@patch_to(Breslow13)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 75
@patch_to(Breslow13)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
    if trajectory is None:
        trajectory = self.trajectory
    else: 
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    if time is None:
        x = trajectory.states[:, 0]
        y = -1.0 * trajectory.states[:, 1]
    else:
        if not isinstance(time, (float, int)):
            raise ValueError("time must be a float or an int")
        else:
            state = trajectory(time)
            x = state[0] 
            y = -1.0 * state[1]
    return np.sqrt(x**2 + y**2)

# %% ../nbs/api/00_models.ipynb 76
@patch_to(Breslow13)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the corrected minimum of x"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    inverted_x = -1 * trajectory.states[:,0]
    cbt_min_idxs, _ = find_peaks(inverted_x)
    cbtmin_times = trajectory.time[cbt_min_idxs] + self.phi_ref
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 77
@patch_to(Breslow13)
def dlmos(self,
          trajectory: DynamicalTrajectory=None # trajectory to calculate the dlmo. If None, the current trajectory is used
          ) -> np.ndarray:
    "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    return self.cbt(trajectory) - self.cbt_to_dlmo

# %% ../nbs/api/00_models.ipynb 80
class Skeldon23(CircadianModel):
    "Implementation of Skeldon's 2023 model from the article 'Method to determine whether sleep phenotypes are driven by endogenous circadian rhythms or environmental light by combining longitudinal data and personalised mathematical models'"
    def __init__(self, params=None):
        default_params = {
            # sleep/wake regulation parameters
            'mu': 17.78, 'chi': 45.0, 'H0': 13.0, 'Delta': 1.0, 'ca': 1.72,
            # light/circadian parameters
            'tauc': 24.2, 'f': 0.99669, 'G': 19.9, 'p': 0.6, 'k': 0.55, 'b': 0.4,
            'gamma': 0.23, 'alpha_0': 0.16, 'beta': 0.013, 'I0': 9500.0,
            'kappa': 24.0 / (2.0 * np.pi),
            # circadian modulation of wakefulness parameters
            'c20': 0.7896, 'alpha21': -0.3912, 'alpha22': 0.7583,
            'beta21': -0.4442, 'beta22': 0.0250, 'beta23': -0.9647, 
            # initial sleep state
            'S0': 0.0,
            # dlmo calculation
            'cbt_to_dlmo': 7.0,
            # forced wake up
            'forced_wakeup_light_threshold': None,
            'forced_wakeup_weekday_only': False,
            }
        num_states = 4 # x, xc, n, H
        num_inputs = 1 # light
        default_initial_condition = np.array([0.23995682, -1.1547196, 0.50529415, 12.83846474]) # condition at midnight for 16L, 8D schedule, S0 = 0
        super(Skeldon23, self).__init__(default_params, num_states, num_inputs, default_initial_condition)
        if params is not None:
            self.parameters = params
            if 'S0' in params:
                self.current_sleep_state = params['S0'] # 0 for wake, 1 for sleep
            else:
                self.current_sleep_state = default_params['S0']
            if 'forced_wakeup_light_threshold' in params:
                self.forced_wakeup_by_light = params['forced_wakeup_light_threshold'] is not None
                self.forced_wakeup_light_threshold = params['forced_wakeup_light_threshold']
            else:
                self.forced_wakeup_by_light = default_params['forced_wakeup_light_threshold'] is not None
                self.forced_wakeup_light_threshold = default_params['forced_wakeup_light_threshold']
            if 'forced_wakeup_weekday_only' in params:
                self.forced_wakeup_weekday_only = params['forced_wakeup_weekday_only']
            else:
                self.forced_wakeup_weekday_only = default_params['forced_wakeup_weekday_only']
        else:
            self.current_sleep_state = default_params['S0'] # 0 for wake, 1 for sleep
            self.forced_wakeup_by_light = default_params['forced_wakeup_light_threshold'] is not None
            self.forced_wakeup_light_threshold = default_params['forced_wakeup_light_threshold']
            self.forced_wakeup_weekday_only = default_params['forced_wakeup_weekday_only']
        # sleep/wake
        self.sleep_state = np.array([self.current_sleep_state])
        # light received by the retina (in lux)
        self.received_light = np.array([])

    def integrate(self,
                  time: np.ndarray, # time points for integration. Time difference between each consecutive pair of values determines step size of the solver
                  initial_condition: np.ndarray=None, # initial state of the model
                  input: np.ndarray=None, # model input (such as light or wake) for each time point
                  ) -> DynamicalTrajectory:
        "Solve the model for specific timepoints given initial conditions and model inputs"
        # input checking
        if input is not None:
            _light_input_checking(input)
        # add additional warning if input has batch dimension
        if initial_condition is not None and initial_condition.ndim >= 2:
            raise ValueError("Skeldon23 model can't be run in batch mode. Please provide a single initial condition")
        # reset sleep state and received light
        self.sleep_state = np.array([self.current_sleep_state])
        self.received_light = np.array([input[0]])
        return super().integrate(time, initial_condition, input)

    def step_rk4(self,
                t: float, # time
                state: np.ndarray, # dynamical state of the model
                input: np.ndarray, # inputs to the model such as light or wake state
                dt: float, # step size in hours 
                ) -> np.ndarray:
        "Integrate the state of the model for one timestep using a fourth-order Runge-Kutta algorithm. Assumes a constant light value for the time step"
        # store received light
        self.received_light = np.append(self.received_light, (1.0 - self.current_sleep_state) * input)
        # calculate the derivative
        new_state = super().step_rk4(t, state, input, dt)
        # calculate the new sleep state
        x = state[0,...]
        xc = state[1,...]
        H = state[3,...]
        S = self.current_sleep_state
        new_sleep_state = S
        linear_term = self.c20 + self.alpha21 * xc + self.alpha22 * x
        quadratic_term = self.beta21 * xc * xc + self.beta22 * xc * x + self.beta23 * x * x
        C = linear_term + quadratic_term # circadian modulation of wakefulness
        if S == 0.0:
            # going from wake to sleep
            H_plus = self.H0 + 0.5 * self.Delta + self.ca * C
            if H >= H_plus:
                new_sleep_state = 1.0
        elif S == 1.0:
            # going from sleep to wake
            H_minus = self.H0 - 0.5 * self.Delta + self.ca * C
            if H <= H_minus:
                new_sleep_state = 0.0
        else:
            raise ValueError("current sleep state must be 0 or 1")
        if self.forced_wakeup_by_light:
            # if light is above threshold, wake up
            if self.forced_wakeup_weekday_only:
                is_weekday = (t / 24.0) % 7 < 5
            else:
                is_weekday = True
            if (input > self.forced_wakeup_light_threshold) and is_weekday:
                new_sleep_state = 0.0
        # update current sleep state and add value to sleep state array
        self.current_sleep_state = new_sleep_state 
        self.sleep_state = np.append(self.sleep_state, new_sleep_state)

        return new_state

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "Skeldon23"

# %% ../nbs/api/00_models.ipynb 81
@patch_to(Skeldon23)
def derv(self, 
         t: float, # time
         state: np.ndarray, # dynamical state (y, x, n, H)
         input: float # light intensity in lux
         ) -> np.ndarray: # derivative of the state
     "Right-hand-side of the differential equation model"
     x = state[0,...]
     xc = state[1,...]
     n = state[2,...]
     H = state[3,...]
     S = self.current_sleep_state
     light = (1.0 - S) * input

     # circadian model derivative
     alpha = self.alpha_0 * pow((light / self.I0), self.p)
     B = self.G * (1.0 - n) * alpha * (1.0 - self.b * x) * (1.0 - self.b * xc)
     gamma_term = self.gamma * (xc - 4.0 / 3.0 * pow(xc, 3.0))
     tauc_term = pow(24.0 / (self.f * self.tauc), 2.0) + self.k * B

     dydt = np.zeros_like(state)
     dydt[0,...] = (1.0 / self.kappa) * (xc + B)
     dydt[1,...] = (1.0 / self.kappa) * (gamma_term - x * tauc_term)
     dydt[2,...] = 60.0 * (alpha * (1.0 - n) - self.beta * n)

     # sleep pressure signal (H(t)) derivative
     dydt[3,...] = (1.0 / self.chi) * (- H + (1.0 - S) * self.mu)


     return dydt

# %% ../nbs/api/00_models.ipynb 82
@patch_to(Skeldon23)
def phase(self,
          trajectory: DynamicalTrajectory=None, # trajectory to calculate the phase. If None, the current trajectory is used
          time: float=None # a time point to calculate the phase at. If None, the phase is calculated for the entire trajectory
          ) -> float:
        if trajectory is None:
            trajectory = self.trajectory
        else: 
            if not isinstance(trajectory, DynamicalTrajectory):
                raise ValueError("trajectory must be a DynamicalTrajectory")
        if time is None:
            x = trajectory.states[:, 0]
            y = -1.0 * trajectory.states[:, 1]
        else:
            if not isinstance(time, (float, int)):
                raise ValueError("time must be a float or an int")
            else:
                state = trajectory(time)
                x = state[0] 
                y = -1.0 * state[1]
        return np.angle(x + complex(0,1) * y)

# %% ../nbs/api/00_models.ipynb 83
@patch_to(Skeldon23)
def amplitude(self,
              trajectory: DynamicalTrajectory=None, # trajectory to calculate the amplitude. If None, the current trajectory is used
              time: float=None # a time point to calculate the amplitude at. If None, the amplitude is calculated for the entire trajectory
              ) -> float:
        if trajectory is None:
            trajectory = self.trajectory
        else: 
            if not isinstance(trajectory, DynamicalTrajectory):
                raise ValueError("trajectory must be a DynamicalTrajectory")
        if time is None:
            x = trajectory.states[:, 0]
            y = -1.0 * trajectory.states[:, 1]
        else:
            if not isinstance(time, (float, int)):
                raise ValueError("time must be a float or an int")
            else:
                state = trajectory(time)
                x = state[0] 
                y = -1.0 * state[1]
        return np.sqrt(x**2 + y**2)

# %% ../nbs/api/00_models.ipynb 84
@patch_to(Skeldon23)
def cbt(self,
        trajectory: DynamicalTrajectory=None, # trajectory to calculate the cbt. If None, the current trajectory is used
        ) -> np.ndarray:
    "Finds the core body temperature minumum markers for the model along a trajectory as the minimum of x"
    if trajectory is None:
        trajectory = self.trajectory
    else:
        if not isinstance(trajectory, DynamicalTrajectory):
            raise ValueError("trajectory must be a DynamicalTrajectory")
    dt = np.diff(trajectory.time)[0]
    inverted_x = -1 * trajectory.states[:,0]
    cbt_min_idxs, _ = find_peaks(inverted_x, distance=np.ceil(self._min_marker_distance_in_hours / dt)) # min separation between troughs
    cbtmin_times = trajectory.time[cbt_min_idxs]
    _check_cbtmin_spacing(cbtmin_times)
    return cbtmin_times

# %% ../nbs/api/00_models.ipynb 85
@patch_to(Skeldon23)
def dlmos(self,
            trajectory: DynamicalTrajectory=None # trajectory to calculate the dlmo. If None, the current trajectory is used
            ) -> np.ndarray:
        "Finds the Dim Light Melatonin Onset (DLMO) markers for the model along a trajectory"
        if trajectory is None:
            trajectory = self.trajectory
        else:
            if not isinstance(trajectory, DynamicalTrajectory):
                raise ValueError("trajectory must be a DynamicalTrajectory")
        return self.cbt(trajectory) - self.cbt_to_dlmo



"""Under construction: Functions to predict sleep"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/06_sleep.ipynb.

# %% auto 0
__all__ = ['TwoProcessModel', 'sleep_midpoint', 'cluster_sleep_periods_scipy']

# %% ../nbs/api/06_sleep.ipynb 4
from math import *
import numpy as np
import pylab as plt
from .plots import *
from .utils import *
from .models import * 
from .lights import *
from scipy.optimize import minimize

# %% ../nbs/api/06_sleep.ipynb 6
class TwoProcessModel:
    "Implementation of the two-process model of sleep regulation."
    def __init__(self, steps_wake_threshold: float = 10.0):
        self.steps_wake_threshold = steps_wake_threshold
        self.awake = True

    @staticmethod
    def check_wake_status(awake: bool, # current wake status
                          h: float,  # homeostat value
                          phase: float #circadian phase
                          ) -> bool:

        c = np.cos(phase)
        H_minus = 0.17
        H_plus = 0.6
        homeostat_a = 0.10

        upper = (H_plus + homeostat_a * c)
        lower = (H_minus + homeostat_a * c)
        above_threshold = h > upper
        below_threshold = h <= lower

        if above_threshold:
            return False
        elif below_threshold:
            return True
        else:
            return awake

    def dhomeostat(self, 
                   homeostat: float, # homeostat value
                   steps: float, # steps value
                   phase: float, #circadian phase
                   ) -> float:

        h = homeostat[0]
        tau_s = 4.2  # hours
        tau_w = 18.2  # hours
        mu_s = 1
        
        step_awake = (steps > self.steps_wake_threshold) or self.awake
        dH = (mu_s - h) / tau_w if step_awake else -h / tau_s
        self.awake = TwoProcessModel.check_wake_status(
            self.awake, 
            h, 
            phase
            )
        return np.array([dH]) 
    
    # implement a RK4 solver
    def step_rk4(self,
                 state: np.ndarray,
                 steps: float,
                 phase: float,
                 dt=0.10):

        k1 = self.dhomeostat(state, steps, phase) 
        k2 = self.dhomeostat(state + k1 * dt / 2.0, steps, phase)
        k3 = self.dhomeostat(state + k2 * dt / 2.0, steps, phase)
        k4 = self.dhomeostat(state + k3 * dt, steps, phase)
        state = state + (dt / 6.0) * (k1 + 2.0*k2 + 2.0*k3 + k4)
        return state

    def __call__(self, 
                 ts: np.ndarray,
                 phase: np.ndarray,
                 steps: np.ndarray,
                 initial_value: np.array = np.array([0.50]) #  initial value for the homeostat
                 ) -> DynamicalTrajectory:
        sol = np.zeros((len(initial_value), len(ts)))
        current_state = initial_value
        sol[:,0] = current_state
        for idx in range(1,len(ts)):
            current_state = self.step_rk4(current_state, steps[idx], phase[idx])
            sol[:,idx] = current_state
        return(DynamicalTrajectory(ts, np.array(sol.T)))

# %% ../nbs/api/06_sleep.ipynb 8
def sleep_midpoint(timetotal: np.ndarray, 
                   Wake: np.ndarray, 
                   durations=True):
    """
        Given a wearable data frame with a Wake column which takes 
        the values 0, 1, missing this routine will create a sleep phase 
        column which is based on constant phase accumulation between sleep 
        midpoints. 

        The sleep midpoints are found using the criteria than they the median 
        time where 

    """
    sleep_start = []
    sleep_end = []
    awake = Wake[0] > 0.50

    if not awake:
        sleep_start.append(timetotal[1])

    for k in range(1, len(Wake)):
        if (Wake[k] > 0.50 and not awake):
            awake = True
            sleep_end.append(timetotal[k])

        if (Wake[k] <= 0.50 and awake):
            awake = False
            sleep_start.append(timetotal[k])

    if Wake[-1] <= 0.50:
        sleep_end.append(timetotal[-1])

    assert len(sleep_start) == len(sleep_end)
    sleep_midpoints = []
    sleep_durations = []
    for (s1, s2) in zip(sleep_start, sleep_end):
        sleep_midpoints += [(s2-s1)/2+s1]
        sleep_durations += [s2-s1]

    if durations:
        return np.array(sleep_midpoints), np.array(sleep_durations)
    else:
        return np.array(sleep_midpoints)

# %% ../nbs/api/06_sleep.ipynb 9
def cluster_sleep_periods_scipy(wake_data: np.ndarray, 
                                epsilon: float,
                                makeplot: bool = False,
                                max_sleep_clusters=None, 
                                min_sleep_clusters=None):
    """
        Given a binary vector wake_data which gives a prediction for the sleep/wake  
        status and a regularization penalty ε this function will create smoothed 
        sleep-wake periods. This helps as preprocessing to remove erroneous short sleep 
        periods (and wake) which may mess up calculations like the sleep midpoint for 
        the day

        cluster_sleep_periods(wake_data : np.ndarray, epsilon: float, makeplot: bool=False):
    """
    np.nan_to_num(wake_data, 0.50)

    def objective(w):
        return sum(w * (1 - wake_data)) + sum((1 - w)*wake_data) + epsilon*sum((w[1:]-w[0:-1])**2)

    max_clusters = max_sleep_clusters or len(wake_data)

    def constraint1(x):
        return max_clusters-sum((x[1:]-x[0:-1])**2)  # geq 0

    min_clusters = min_sleep_clusters or 0

    print(
        f"The max clusters are {max_clusters} and the min clusters are {min_clusters}")

    def constraint2(x):
        return sum((x[1:]-x[0:-1])**2)-min_clusters  # geq 0

    bnds = (0.0, 1.0)
    all_bnds = [bnds for b in range(len(wake_data))]

    constraint1d = {'type': 'ineq', 'fun': constraint1}
    constraint2d = {'type': 'ineq', 'fun': constraint2}
    all_cons = [constraint1d, constraint2d]

    x0 = wake_data
    sol = minimize(objective, x0, method='SLSQP', bounds=all_bnds)

    if makeplot:
        pl = plt.scatter(range(len(wake_data)), wake_data + 0.1 *
                         np.random.randn(len(wake_data)), label="", color="blue")
        plt.plot(range(len(wake_data)), convert_binary(
            sol.x), lw=2.0, label="", color="red")
        plt.show()

    print(
        f"The max clusters are {max_clusters} takes value {constraint1(sol.x)}>=0.0")
    print(
        f"The min clusters are {min_clusters} and takes the value {constraint2(sol.x)}>=0.0")
    return(convert_binary(sol.x))



"""Defines methods for generating synthetic data with varying levels of realism."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/10_synthetic_data.ipynb.

# %% auto 0
__all__ = ['generate_activity_from_light']

# %% ../nbs/api/10_synthetic_data.ipynb 4
import matplotlib.pyplot as plt
import numpy as np
from .lights import LightSchedule

# %% ../nbs/api/10_synthetic_data.ipynb 5
def generate_activity_from_light(
    time: np.ndarray, # time points.
    light: LightSchedule, # light schedule.
    mu_l: float=5.0, # mean of low activity levels.
    mu_h: float=25.0, # mean of high activity levels.
    mu_s: float=0.0, # mean of sleep activity levels.
    sigma_l: float=7.5, # std of low activity levels.
    sigma_h: float=30.0, # std of high activity levels.
    sigma_s: float=2.0, # std of activity during sleep.
    active_level: float=0.5, # person's activity level from 0 to 1. Represents the probability of choosing from the high or low activity distributions at any point in time. Values close to 1 increase the proportion of high activity values.
) -> np.ndarray:
    "Generates an 'activity' schedule from an input light schedule. The activity schedule produced is a corrupted version of the light schedule."
    # input validation
    if not isinstance(mu_l, (float, int)):
        raise TypeError(f"mu_l must be a float or int, got {type(mu_l)}")
    elif mu_l < 0.0:
        raise ValueError(f"mu_l must be a nonnegative float or int, got {mu_l}")
    if not isinstance(mu_h, (float, int)):
        raise TypeError(f"mu_h must be a float or int, got {type(mu_h)}")
    elif mu_h < 0.0:
        raise ValueError(f"mu_h must be a nonnegative float or int, got {mu_h}")
    if not isinstance(mu_s, (float, int)):
        raise TypeError(f"mu_s must be a float or int, got {type(mu_s)}")
    elif mu_s < 0.0:
        raise ValueError(f"mu_s must be a nonnegative float or int, got {mu_s}")
    if not isinstance(sigma_l, (float, int)):
        raise TypeError(f"sigma_l must be a float or int, got {type(sigma_l)}")
    elif sigma_l < 0.0:
        raise ValueError(f"sigma_l must be a nonnegative float or int, got {sigma_l}")
    if not isinstance(sigma_h, (float, int)):
        raise TypeError(f"sigma_h must be a float or int, got {type(sigma_h)}")
    elif sigma_h < 0.0:
        raise ValueError(f"sigma_h must be a nonnegative float or int, got {sigma_h}")
    if not isinstance(sigma_s, (float, int)):
        raise TypeError(f"sigma_s must be a float or int, got {type(sigma_s)}")
    elif sigma_s < 0.0:
        raise ValueError(f"sigma_s must be a nonnegative float or int, got {sigma_s}")
    if not isinstance(active_level, (float, int)):
        raise TypeError(f"active_level must be a float or int, got {type(active_level)}")
    elif active_level < 0.0 or active_level > 1.0:
        raise ValueError(f"active_level must be between 0.0 and 1.0, got {active_level}")
    # check that time is a Numpy array
    if not isinstance(time, np.ndarray):
        raise TypeError("`time` must be a Numpy array")
    # check that the light schedule is a LightSchedule
    if not isinstance(light, LightSchedule):
        raise TypeError("`light` must be a `LightSchedule` object")

    # create the activity vector the same size as time
    activity = np.zeros(time.shape)
    for t in range(time.shape[0]):
        if light(time[t]) == 0:
            activity[t] = np.abs(np.random.normal(mu_s, sigma_s))
        elif light(time[t]) > 0:
            high_activity = np.random.uniform() <= active_level
            if high_activity == True:
                activity[t] = np.abs(np.random.normal(mu_h, sigma_h))
            elif high_activity == False:
                activity[t] = np.abs(np.random.normal(mu_l, sigma_l))
    return activity



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/circadian',
                'doc_host': 'https://arcascope.github.io/circadian/',
                'git_url': 'https://github.com/Arcascope/circadian',
                'lib_path': 'circadian'},
  'syms': { 'circadian.cli': { 'circadian.cli.main_acto': ('api/cli.html#main_acto', 'circadian/cli.py'),
                               'circadian.cli.main_esri': ('api/cli.html#main_esri', 'circadian/cli.py')},
            'circadian.lights': { 'circadian.lights.LightSchedule': ('api/lights.html#lightschedule', 'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.Chang14': ( 'api/lights.html#lightschedule.chang14',
                                                                              'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.Hilaire12': ( 'api/lights.html#lightschedule.hilaire12',
                                                                                'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.Regular': ( 'api/lights.html#lightschedule.regular',
                                                                              'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.School': ('api/lights.html#lightschedule.school', 'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.ShiftWork': ( 'api/lights.html#lightschedule.shiftwork',
                                                                                'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.SlamShift': ( 'api/lights.html#lightschedule.slamshift',
                                                                                'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.SocialJetlag': ( 'api/lights.html#lightschedule.socialjetlag',
                                                                                   'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.__add__': ( 'api/lights.html#lightschedule.__add__',
                                                                              'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.__call__': ( 'api/lights.html#lightschedule.__call__',
                                                                               'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.__init__': ( 'api/lights.html#lightschedule.__init__',
                                                                               'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.__sub__': ( 'api/lights.html#lightschedule.__sub__',
                                                                              'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.concatenate_at': ( 'api/lights.html#lightschedule.concatenate_at',
                                                                                     'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.from_pulse': ( 'api/lights.html#lightschedule.from_pulse',
                                                                                 'circadian/lights.py'),
                                  'circadian.lights.LightSchedule.plot': ('api/lights.html#lightschedule.plot', 'circadian/lights.py'),
                                  'circadian.lights.outdoor_light': ('api/lights.html#outdoor_light', 'circadian/lights.py')},
            'circadian.metrics': {'circadian.metrics.esri': ('api/metrics.html#esri', 'circadian/metrics.py')},
            'circadian.models': { 'circadian.models.Breslow13': ('api/models.html#breslow13', 'circadian/models.py'),
                                  'circadian.models.Breslow13.__init__': ('api/models.html#breslow13.__init__', 'circadian/models.py'),
                                  'circadian.models.Breslow13.__repr__': ('api/models.html#breslow13.__repr__', 'circadian/models.py'),
                                  'circadian.models.Breslow13.__str__': ('api/models.html#breslow13.__str__', 'circadian/models.py'),
                                  'circadian.models.Breslow13.amplitude': ('api/models.html#breslow13.amplitude', 'circadian/models.py'),
                                  'circadian.models.Breslow13.avoid_negative_H1': ( 'api/models.html#breslow13.avoid_negative_h1',
                                                                                    'circadian/models.py'),
                                  'circadian.models.Breslow13.cbt': ('api/models.html#breslow13.cbt', 'circadian/models.py'),
                                  'circadian.models.Breslow13.derv': ('api/models.html#breslow13.derv', 'circadian/models.py'),
                                  'circadian.models.Breslow13.dlmos': ('api/models.html#breslow13.dlmos', 'circadian/models.py'),
                                  'circadian.models.Breslow13.integrate': ('api/models.html#breslow13.integrate', 'circadian/models.py'),
                                  'circadian.models.Breslow13.melatonin_drive': ( 'api/models.html#breslow13.melatonin_drive',
                                                                                  'circadian/models.py'),
                                  'circadian.models.Breslow13.phase': ('api/models.html#breslow13.phase', 'circadian/models.py'),
                                  'circadian.models.CircadianModel': ('api/models.html#circadianmodel', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.__call__': ( 'api/models.html#circadianmodel.__call__',
                                                                                'circadian/models.py'),
                                  'circadian.models.CircadianModel.__init__': ( 'api/models.html#circadianmodel.__init__',
                                                                                'circadian/models.py'),
                                  'circadian.models.CircadianModel._default_initial_condition': ( 'api/models.html#circadianmodel._default_initial_condition',
                                                                                                  'circadian/models.py'),
                                  'circadian.models.CircadianModel._default_params': ( 'api/models.html#circadianmodel._default_params',
                                                                                       'circadian/models.py'),
                                  'circadian.models.CircadianModel._num_inputs': ( 'api/models.html#circadianmodel._num_inputs',
                                                                                   'circadian/models.py'),
                                  'circadian.models.CircadianModel._num_states': ( 'api/models.html#circadianmodel._num_states',
                                                                                   'circadian/models.py'),
                                  'circadian.models.CircadianModel.amplitude': ( 'api/models.html#circadianmodel.amplitude',
                                                                                 'circadian/models.py'),
                                  'circadian.models.CircadianModel.cbt': ('api/models.html#circadianmodel.cbt', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.derv': ('api/models.html#circadianmodel.derv', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.dlmos': ('api/models.html#circadianmodel.dlmos', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.equilibrate': ( 'api/models.html#circadianmodel.equilibrate',
                                                                                   'circadian/models.py'),
                                  'circadian.models.CircadianModel.get_parameters_array': ( 'api/models.html#circadianmodel.get_parameters_array',
                                                                                            'circadian/models.py'),
                                  'circadian.models.CircadianModel.initial_condition': ( 'api/models.html#circadianmodel.initial_condition',
                                                                                         'circadian/models.py'),
                                  'circadian.models.CircadianModel.integrate': ( 'api/models.html#circadianmodel.integrate',
                                                                                 'circadian/models.py'),
                                  'circadian.models.CircadianModel.parameters': ( 'api/models.html#circadianmodel.parameters',
                                                                                  'circadian/models.py'),
                                  'circadian.models.CircadianModel.phase': ('api/models.html#circadianmodel.phase', 'circadian/models.py'),
                                  'circadian.models.CircadianModel.step_rk4': ( 'api/models.html#circadianmodel.step_rk4',
                                                                                'circadian/models.py'),
                                  'circadian.models.CircadianModel.trajectory': ( 'api/models.html#circadianmodel.trajectory',
                                                                                  'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory': ('api/models.html#dynamicaltrajectory', 'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__call__': ( 'api/models.html#dynamicaltrajectory.__call__',
                                                                                     'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__getitem__': ( 'api/models.html#dynamicaltrajectory.__getitem__',
                                                                                        'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__init__': ( 'api/models.html#dynamicaltrajectory.__init__',
                                                                                     'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__len__': ( 'api/models.html#dynamicaltrajectory.__len__',
                                                                                    'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.__str__': ( 'api/models.html#dynamicaltrajectory.__str__',
                                                                                    'circadian/models.py'),
                                  'circadian.models.DynamicalTrajectory.get_batch': ( 'api/models.html#dynamicaltrajectory.get_batch',
                                                                                      'circadian/models.py'),
                                  'circadian.models.Forger99': ('api/models.html#forger99', 'circadian/models.py'),
                                  'circadian.models.Forger99.__init__': ('api/models.html#forger99.__init__', 'circadian/models.py'),
                                  'circadian.models.Forger99.__repr__': ('api/models.html#forger99.__repr__', 'circadian/models.py'),
                                  'circadian.models.Forger99.__str__': ('api/models.html#forger99.__str__', 'circadian/models.py'),
                                  'circadian.models.Forger99.amplitude': ('api/models.html#forger99.amplitude', 'circadian/models.py'),
                                  'circadian.models.Forger99.cbt': ('api/models.html#forger99.cbt', 'circadian/models.py'),
                                  'circadian.models.Forger99.derv': ('api/models.html#forger99.derv', 'circadian/models.py'),
                                  'circadian.models.Forger99.dlmos': ('api/models.html#forger99.dlmos', 'circadian/models.py'),
                                  'circadian.models.Forger99.integrate': ('api/models.html#forger99.integrate', 'circadian/models.py'),
                                  'circadian.models.Forger99.phase': ('api/models.html#forger99.phase', 'circadian/models.py'),
                                  'circadian.models.Hannay19': ('api/models.html#hannay19', 'circadian/models.py'),
                                  'circadian.models.Hannay19.__init__': ('api/models.html#hannay19.__init__', 'circadian/models.py'),
                                  'circadian.models.Hannay19.__repr__': ('api/models.html#hannay19.__repr__', 'circadian/models.py'),
                                  'circadian.models.Hannay19.__str__': ('api/models.html#hannay19.__str__', 'circadian/models.py'),
                                  'circadian.models.Hannay19.amplitude': ('api/models.html#hannay19.amplitude', 'circadian/models.py'),
                                  'circadian.models.Hannay19.cbt': ('api/models.html#hannay19.cbt', 'circadian/models.py'),
                                  'circadian.models.Hannay19.derv': ('api/models.html#hannay19.derv', 'circadian/models.py'),
                                  'circadian.models.Hannay19.dlmos': ('api/models.html#hannay19.dlmos', 'circadian/models.py'),
                                  'circadian.models.Hannay19.integrate': ('api/models.html#hannay19.integrate', 'circadian/models.py'),
                                  'circadian.models.Hannay19.phase': ('api/models.html#hannay19.phase', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP': ('api/models.html#hannay19tp', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.__init__': ('api/models.html#hannay19tp.__init__', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.__repr__': ('api/models.html#hannay19tp.__repr__', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.__str__': ('api/models.html#hannay19tp.__str__', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.amplitude': ('api/models.html#hannay19tp.amplitude', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.cbt': ('api/models.html#hannay19tp.cbt', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.derv': ('api/models.html#hannay19tp.derv', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.dlmos': ('api/models.html#hannay19tp.dlmos', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.integrate': ('api/models.html#hannay19tp.integrate', 'circadian/models.py'),
                                  'circadian.models.Hannay19TP.phase': ('api/models.html#hannay19tp.phase', 'circadian/models.py'),
                                  'circadian.models.Hilaire07': ('api/models.html#hilaire07', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.__init__': ('api/models.html#hilaire07.__init__', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.__repr__': ('api/models.html#hilaire07.__repr__', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.__str__': ('api/models.html#hilaire07.__str__', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.amplitude': ('api/models.html#hilaire07.amplitude', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.cbt': ('api/models.html#hilaire07.cbt', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.derv': ('api/models.html#hilaire07.derv', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.dlmos': ('api/models.html#hilaire07.dlmos', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.integrate': ('api/models.html#hilaire07.integrate', 'circadian/models.py'),
                                  'circadian.models.Hilaire07.phase': ('api/models.html#hilaire07.phase', 'circadian/models.py'),
                                  'circadian.models.Jewett99': ('api/models.html#jewett99', 'circadian/models.py'),
                                  'circadian.models.Jewett99.__init__': ('api/models.html#jewett99.__init__', 'circadian/models.py'),
                                  'circadian.models.Jewett99.__repr__': ('api/models.html#jewett99.__repr__', 'circadian/models.py'),
                                  'circadian.models.Jewett99.__str__': ('api/models.html#jewett99.__str__', 'circadian/models.py'),
                                  'circadian.models.Jewett99.amplitude': ('api/models.html#jewett99.amplitude', 'circadian/models.py'),
                                  'circadian.models.Jewett99.cbt': ('api/models.html#jewett99.cbt', 'circadian/models.py'),
                                  'circadian.models.Jewett99.derv': ('api/models.html#jewett99.derv', 'circadian/models.py'),
                                  'circadian.models.Jewett99.dlmos': ('api/models.html#jewett99.dlmos', 'circadian/models.py'),
                                  'circadian.models.Jewett99.integrate': ('api/models.html#jewett99.integrate', 'circadian/models.py'),
                                  'circadian.models.Jewett99.phase': ('api/models.html#jewett99.phase', 'circadian/models.py'),
                                  'circadian.models.Skeldon23': ('api/models.html#skeldon23', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.__init__': ('api/models.html#skeldon23.__init__', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.__repr__': ('api/models.html#skeldon23.__repr__', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.__str__': ('api/models.html#skeldon23.__str__', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.amplitude': ('api/models.html#skeldon23.amplitude', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.cbt': ('api/models.html#skeldon23.cbt', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.derv': ('api/models.html#skeldon23.derv', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.dlmos': ('api/models.html#skeldon23.dlmos', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.integrate': ('api/models.html#skeldon23.integrate', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.phase': ('api/models.html#skeldon23.phase', 'circadian/models.py'),
                                  'circadian.models.Skeldon23.step_rk4': ('api/models.html#skeldon23.step_rk4', 'circadian/models.py'),
                                  'circadian.models._check_cbtmin_spacing': ( 'api/models.html#_check_cbtmin_spacing',
                                                                              'circadian/models.py'),
                                  'circadian.models._get_default_initial_condition': ( 'api/models.html#_get_default_initial_condition',
                                                                                       'circadian/models.py'),
                                  'circadian.models._initial_condition_input_checking': ( 'api/models.html#_initial_condition_input_checking',
                                                                                          'circadian/models.py'),
                                  'circadian.models._light_input_checking': ( 'api/models.html#_light_input_checking',
                                                                              'circadian/models.py'),
                                  'circadian.models._model_input_checking': ( 'api/models.html#_model_input_checking',
                                                                              'circadian/models.py'),
                                  'circadian.models._parameter_input_checking': ( 'api/models.html#_parameter_input_checking',
                                                                                  'circadian/models.py'),
                                  'circadian.models._positive_int_checking': ( 'api/models.html#_positive_int_checking',
                                                                               'circadian/models.py'),
                                  'circadian.models._state_input_checking': ( 'api/models.html#_state_input_checking',
                                                                              'circadian/models.py'),
                                  'circadian.models._time_input_checking': ('api/models.html#_time_input_checking', 'circadian/models.py'),
                                  'circadian.models._wake_input_checking': ('api/models.html#_wake_input_checking', 'circadian/models.py')},
            'circadian.phasetools': { 'circadian.phasetools.cosinor': ('api/phasetools.html#cosinor', 'circadian/phasetools.py'),
                                      'circadian.phasetools.cosinor_goals': ( 'api/phasetools.html#cosinor_goals',
                                                                              'circadian/phasetools.py'),
                                      'circadian.phasetools.cosinor_phase': ( 'api/phasetools.html#cosinor_phase',
                                                                              'circadian/phasetools.py')},
            'circadian.plots': { 'circadian.plots.Actogram': ('api/plots.html#actogram', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.__init__': ('api/plots.html#actogram.__init__', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.addLightSchedule': ( 'api/plots.html#actogram.addlightschedule',
                                                                                'circadian/plots.py'),
                                 'circadian.plots.Actogram.addRect': ('api/plots.html#actogram.addrect', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.getRectangles': ('api/plots.html#actogram.getrectangles', 'circadian/plots.py'),
                                 'circadian.plots.Actogram.plot_phasemarker': ( 'api/plots.html#actogram.plot_phasemarker',
                                                                                'circadian/plots.py'),
                                 'circadian.plots.Actogram.plot_phasetimes': ( 'api/plots.html#actogram.plot_phasetimes',
                                                                               'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic': ('api/plots.html#stroboscopic', 'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic.__init__': ('api/plots.html#stroboscopic.__init__', 'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic._make_strobo_plot': ( 'api/plots.html#stroboscopic._make_strobo_plot',
                                                                                     'circadian/plots.py'),
                                 'circadian.plots.Stroboscopic.add_strobo_plot': ( 'api/plots.html#stroboscopic.add_strobo_plot',
                                                                                   'circadian/plots.py'),
                                 'circadian.plots.plot_actogram': ('api/plots.html#plot_actogram', 'circadian/plots.py'),
                                 'circadian.plots.plot_mae': ('api/plots.html#plot_mae', 'circadian/plots.py'),
                                 'circadian.plots.plot_torus': ('api/plots.html#plot_torus', 'circadian/plots.py')},
            'circadian.prc': { 'circadian.prc.DosageResponseCurve': ('api/prc.html#dosageresponsecurve', 'circadian/prc.py'),
                               'circadian.prc.DosageResponseCurve.__init__': ( 'api/prc.html#dosageresponsecurve.__init__',
                                                                               'circadian/prc.py'),
                               'circadian.prc.DosageResponseCurve.light_dosage': ( 'api/prc.html#dosageresponsecurve.light_dosage',
                                                                                   'circadian/prc.py'),
                               'circadian.prc.DosageResponseCurve.light_dosage_day1': ( 'api/prc.html#dosageresponsecurve.light_dosage_day1',
                                                                                        'circadian/prc.py'),
                               'circadian.prc.IntensityResponseCurveLight': ( 'api/prc.html#intensityresponsecurvelight',
                                                                              'circadian/prc.py'),
                               'circadian.prc.IntensityResponseCurveLight.__init__': ( 'api/prc.html#intensityresponsecurvelight.__init__',
                                                                                       'circadian/prc.py'),
                               'circadian.prc.IntensityResponseCurveLight.light_intensity': ( 'api/prc.html#intensityresponsecurvelight.light_intensity',
                                                                                              'circadian/prc.py'),
                               'circadian.prc.PRCFinder': ('api/prc.html#prcfinder', 'circadian/prc.py'),
                               'circadian.prc.PRCFinder.exp_type0': ('api/prc.html#prcfinder.exp_type0', 'circadian/prc.py'),
                               'circadian.prc.PRCFinder.prc_type0_point': ('api/prc.html#prcfinder.prc_type0_point', 'circadian/prc.py'),
                               'circadian.prc.PRCFinder.type0x': ('api/prc.html#prcfinder.type0x', 'circadian/prc.py'),
                               'circadian.prc.PhaseResponseCurveLight': ('api/prc.html#phaseresponsecurvelight', 'circadian/prc.py'),
                               'circadian.prc.PhaseResponseCurveLight.__init__': ( 'api/prc.html#phaseresponsecurvelight.__init__',
                                                                                   'circadian/prc.py'),
                               'circadian.prc.PhaseResponseCurveLight.light_amplitude_resetting': ( 'api/prc.html#phaseresponsecurvelight.light_amplitude_resetting',
                                                                                                    'circadian/prc.py'),
                               'circadian.prc.PhaseResponseCurveLight.light_czeisler_type0': ( 'api/prc.html#phaseresponsecurvelight.light_czeisler_type0',
                                                                                               'circadian/prc.py'),
                               'circadian.prc.PhaseResponseCurveLight.light_hilaire': ( 'api/prc.html#phaseresponsecurvelight.light_hilaire',
                                                                                        'circadian/prc.py'),
                               'circadian.prc.PhaseResponseCurveLight.light_khalsa': ( 'api/prc.html#phaseresponsecurvelight.light_khalsa',
                                                                                       'circadian/prc.py'),
                               'circadian.prc.RimmerLightPulseLight': ('api/prc.html#rimmerlightpulselight', 'circadian/prc.py'),
                               'circadian.prc.RimmerLightPulseLight.__init__': ( 'api/prc.html#rimmerlightpulselight.__init__',
                                                                                 'circadian/prc.py'),
                               'circadian.prc.RimmerLightPulseLight.make_pulse_rimmer': ( 'api/prc.html#rimmerlightpulselight.make_pulse_rimmer',
                                                                                          'circadian/prc.py'),
                               'circadian.prc.RimmerLightPulseLight.pulse_rimmer_start': ( 'api/prc.html#rimmerlightpulselight.pulse_rimmer_start',
                                                                                           'circadian/prc.py'),
                               'circadian.prc.get_pulse': ('api/prc.html#get_pulse', 'circadian/prc.py'),
                               'circadian.prc.heaviside': ('api/prc.html#heaviside', 'circadian/prc.py'),
                               'circadian.prc.make_pulse': ('api/prc.html#make_pulse', 'circadian/prc.py')},
            'circadian.readers': { 'circadian.readers.WearableData': ('api/readers.html#wearabledata', 'circadian/readers.py'),
                                   'circadian.readers.WearableData.__init__': ( 'api/readers.html#wearabledata.__init__',
                                                                                'circadian/readers.py'),
                                   'circadian.readers.WearableData._validate_columns': ( 'api/readers.html#wearabledata._validate_columns',
                                                                                         'circadian/readers.py'),
                                   'circadian.readers.WearableData._validate_metadata': ( 'api/readers.html#wearabledata._validate_metadata',
                                                                                          'circadian/readers.py'),
                                   'circadian.readers.WearableData.add_metadata': ( 'api/readers.html#wearabledata.add_metadata',
                                                                                    'circadian/readers.py'),
                                   'circadian.readers.WearableData.is_valid': ( 'api/readers.html#wearabledata.is_valid',
                                                                                'circadian/readers.py'),
                                   'circadian.readers.WearableData.rename_columns': ( 'api/readers.html#wearabledata.rename_columns',
                                                                                      'circadian/readers.py'),
                                   'circadian.readers.combine_wearable_dataframes': ( 'api/readers.html#combine_wearable_dataframes',
                                                                                      'circadian/readers.py'),
                                   'circadian.readers.interval_fraction': ('api/readers.html#interval_fraction', 'circadian/readers.py'),
                                   'circadian.readers.load_actiwatch': ('api/readers.html#load_actiwatch', 'circadian/readers.py'),
                                   'circadian.readers.load_csv': ('api/readers.html#load_csv', 'circadian/readers.py'),
                                   'circadian.readers.load_json': ('api/readers.html#load_json', 'circadian/readers.py'),
                                   'circadian.readers.resample_df': ('api/readers.html#resample_df', 'circadian/readers.py')},
            'circadian.sleep': { 'circadian.sleep.TwoProcessModel': ('api/sleep.html#twoprocessmodel', 'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.__call__': ( 'api/sleep.html#twoprocessmodel.__call__',
                                                                               'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.__init__': ( 'api/sleep.html#twoprocessmodel.__init__',
                                                                               'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.check_wake_status': ( 'api/sleep.html#twoprocessmodel.check_wake_status',
                                                                                        'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.dhomeostat': ( 'api/sleep.html#twoprocessmodel.dhomeostat',
                                                                                 'circadian/sleep.py'),
                                 'circadian.sleep.TwoProcessModel.step_rk4': ( 'api/sleep.html#twoprocessmodel.step_rk4',
                                                                               'circadian/sleep.py'),
                                 'circadian.sleep.cluster_sleep_periods_scipy': ( 'api/sleep.html#cluster_sleep_periods_scipy',
                                                                                  'circadian/sleep.py'),
                                 'circadian.sleep.sleep_midpoint': ('api/sleep.html#sleep_midpoint', 'circadian/sleep.py')},
            'circadian.synthetic_data': { 'circadian.synthetic_data.generate_activity_from_light': ( 'api/synthetic_data.html#generate_activity_from_light',
                                                                                                     'circadian/synthetic_data.py')},
            'circadian.utils': { 'circadian.utils.abs_hour_diff': ('api/utils.html#abs_hour_diff', 'circadian/utils.py'),
                                 'circadian.utils.amplitude_percent_change': ( 'api/utils.html#amplitude_percent_change',
                                                                               'circadian/utils.py'),
                                 'circadian.utils.angle_difference': ('api/utils.html#angle_difference', 'circadian/utils.py'),
                                 'circadian.utils.cal_days_diff': ('api/utils.html#cal_days_diff', 'circadian/utils.py'),
                                 'circadian.utils.convert_binary': ('api/utils.html#convert_binary', 'circadian/utils.py'),
                                 'circadian.utils.cut_phases_12': ('api/utils.html#cut_phases_12', 'circadian/utils.py'),
                                 'circadian.utils.parse_dt': ('api/utils.html#parse_dt', 'circadian/utils.py'),
                                 'circadian.utils.phase_coherence': ('api/utils.html#phase_coherence', 'circadian/utils.py'),
                                 'circadian.utils.phase_coherence_clock': ('api/utils.html#phase_coherence_clock', 'circadian/utils.py'),
                                 'circadian.utils.phase_difference': ('api/utils.html#phase_difference', 'circadian/utils.py'),
                                 'circadian.utils.phase_ic_guess': ('api/utils.html#phase_ic_guess', 'circadian/utils.py'),
                                 'circadian.utils.sleep_midpoint_and_duration': ( 'api/utils.html#sleep_midpoint_and_duration',
                                                                                  'circadian/utils.py'),
                                 'circadian.utils.subtract_clock_times': ('api/utils.html#subtract_clock_times', 'circadian/utils.py'),
                                 'circadian.utils.times_to_angle': ('api/utils.html#times_to_angle', 'circadian/utils.py'),
                                 'circadian.utils.utc_to_hrs': ('api/utils.html#utc_to_hrs', 'circadian/utils.py')}}}



"""Defines metrics to quantify circadian disruption"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/04_metrics.ipynb.

# %% auto 0
__all__ = ['esri']

# %% ../nbs/api/04_metrics.ipynb 4
import warnings
import numpy as np
from typing import List
from .models import Hannay19
from .lights import LightSchedule

# %% ../nbs/api/04_metrics.ipynb 5
def esri(time: np.ndarray, # time in hours to use for the simulation 
         light_schedule: np.ndarray, # light schedule in lux 
         analysis_days: int=4, # number of days used to calculate ESRI
         esri_dt: float=1.0, # time resolution of the ESRI calculation in hours
         initial_amplitude: float=0.1, # initial amplitude for the simulation. This is the ESRI value for constant darkness
         phase_at_midnight: float=1.65238233, # phase at midnight. Default value corresponds to a 8 hour darkness and 16 hour light schedule with wake at 8 am.
         ) -> List: # list with ESRI timepoints and ESRI values. Negative ESRI values are turned into NaNs
        "Calculate the ESRI metric for a given light schedule. Follows the implementation from Moreno et al. 2023 'Validation of the Entrainment Signal Regularity Index and associations with children's changes in BMI'"
        # validate inputs
        if not isinstance(time, np.ndarray):
            raise TypeError(f'time must be a numpy array, not {type(time)}')
        if not isinstance(light_schedule, np.ndarray):
            raise TypeError(f'light_schedule must be a numpy array, not {type(light_schedule)}')
        if len(time) != len(light_schedule):
            raise ValueError(f'time and light_schedule must be the same length')
        if not np.all(np.isclose(np.diff(time), np.diff(time)[0])):
            raise ValueError(f'time must have a fixed time resolution (time between timepoints must be constant)')
        if not isinstance(analysis_days, int):
            raise TypeError(f'analysis_days must be an integer, not {type(analysis_days)}')
        if analysis_days < 1:
            raise ValueError(f'analysis_days must be greater than 0')
        if not isinstance(esri_dt, (int, float)):
            raise TypeError(f'esri_dt must be a float or an int, not {type(esri_dt)}')
        if esri_dt <= 0:
            raise ValueError(f'esri_dt must be greater than 0')
        if not isinstance(initial_amplitude, (int, float)):
            raise TypeError(f'initial_amplitude must be a float or an int, not {type(initial_amplitude)}')
        if initial_amplitude < 0:
            raise ValueError(f'initial_amplitude must be non-negative')
        # calculate ESRI 
        model = Hannay19(params={'K': 0.0, 'gamma': 0.0}) # with these parameters, amplitude is constant in the absence of light
        simulation_dt = np.diff(time)[0]
        esri_time = np.arange(time[0], time[-1] - analysis_days*24, esri_dt)
        esri_array = np.zeros_like(esri_time)
        for idx, t in enumerate(esri_time):
            initial_phase = phase_at_midnight + np.mod(t, 24.0) * np.pi / 12 # assumes regular schedule with wake at 8 am
            initial_condition = np.array([initial_amplitude, initial_phase, 0.0])
            simulation_time = np.arange(t, t + analysis_days*24, simulation_dt)
            simulation_light = np.interp(simulation_time, time, light_schedule)
            trajectory = model(simulation_time, initial_condition, simulation_light)
            esri_value = trajectory.states[-1, 0] # model amplitude at the end of the simulation
            esri_array[idx] = esri_value
        # clean up any negative values
        esri_array[esri_array < 0] = np.nan
        # if there's any NaNs, throw a warning thay probably dt was too small
        if np.any(np.isnan(esri_array)):
            warnings.warn(f'ESRI calculation failed for certain timepoints (NaN ESRI values). Try decreasing the time resolution of the `time` and `light_schedule` arrays.')
        return esri_time, esri_array
